# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2013 Matthieu Huin
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Swift (OpenStack Object Storage) backend.
"""

import os

import duplicity.backend
from duplicity import config
from duplicity import errors
from duplicity import log
from duplicity.errors import BackendException

global Connection, ClientException, SwiftService, SwiftUploadObject


class SwiftBackend(duplicity.backend.Backend):
    """
    Backend for Swift
    """

    def __init__(self, parsed_url):
        """
        Initialize the Swift backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        global Connection, ClientException, SwiftService, SwiftUploadObject
        try:
            from swiftclient.service import SwiftUploadObject
            from swiftclient.service import SwiftService
            from swiftclient import Connection
            from swiftclient import ClientException
        except ImportError as e:
            raise BackendException(f"""Swift backend requires the python-swiftclient library.
Exception: {str(e)}""")

        self.resp_exc = ClientException
        conn_kwargs = {}
        os_options = {}
        svc_options = {}

        # if the user has already authenticated
        if "SWIFT_PREAUTHURL" in os.environ and "SWIFT_PREAUTHTOKEN" in os.environ:
            conn_kwargs["preauthurl"] = os.environ["SWIFT_PREAUTHURL"]
            conn_kwargs["preauthtoken"] = os.environ["SWIFT_PREAUTHTOKEN"]

        else:
            if "SWIFT_USERNAME" not in os.environ:
                raise BackendException("SWIFT_USERNAME environment variable " "not set.")

            if "SWIFT_PASSWORD" not in os.environ:
                raise BackendException("SWIFT_PASSWORD environment variable " "not set.")

            if "SWIFT_AUTHURL" not in os.environ:
                raise BackendException("SWIFT_AUTHURL environment variable " "not set.")

            svc_options["os_username"] = conn_kwargs["user"] = os.environ["SWIFT_USERNAME"]
            svc_options["os_password"] = conn_kwargs["key"] = os.environ["SWIFT_PASSWORD"]
            svc_options["os_auth_url"] = conn_kwargs["authurl"] = os.environ["SWIFT_AUTHURL"]

        if "SWIFT_AUTHVERSION" in os.environ:
            svc_options["auth_version"] = conn_kwargs["auth_version"] = os.environ["SWIFT_AUTHVERSION"]
            if os.environ["SWIFT_AUTHVERSION"] == "3":
                if "SWIFT_USER_DOMAIN_NAME" in os.environ:
                    os_options.update({"user_domain_name": os.environ["SWIFT_USER_DOMAIN_NAME"]})
                if "SWIFT_USER_DOMAIN_ID" in os.environ:
                    os_options.update({"user_domain_id": os.environ["SWIFT_USER_DOMAIN_ID"]})
                if "SWIFT_PROJECT_DOMAIN_NAME" in os.environ:
                    os_options.update({"project_domain_name": os.environ["SWIFT_PROJECT_DOMAIN_NAME"]})
                if "SWIFT_PROJECT_DOMAIN_ID" in os.environ:
                    os_options.update({"project_domain_id": os.environ["SWIFT_PROJECT_DOMAIN_ID"]})
                if "SWIFT_PROJECT_ID" in os.environ:
                    os_options.update({"project_id": os.environ["SWIFT_PROJECT_ID"]})
                if "SWIFT_PROJECT_NAME" in os.environ:
                    os_options.update({"project_name": os.environ["SWIFT_PROJECT_NAME"]})
                if "SWIFT_TENANTNAME" in os.environ:
                    os_options.update({"tenant_name": os.environ["SWIFT_TENANTNAME"]})
                if "SWIFT_ENDPOINT_TYPE" in os.environ:
                    os_options.update({"endpoint_type": os.environ["SWIFT_ENDPOINT_TYPE"]})
                if "SWIFT_USERID" in os.environ:
                    os_options.update({"user_id": os.environ["SWIFT_USERID"]})
                if "SWIFT_TENANTID" in os.environ:
                    os_options.update({"tenant_id": os.environ["SWIFT_TENANTID"]})
                if "SWIFT_REGIONNAME" in os.environ:
                    os_options.update({"region_name": os.environ["SWIFT_REGIONNAME"]})

        else:
            conn_kwargs["auth_version"] = "1"
        if "SWIFT_TENANTNAME" in os.environ:
            conn_kwargs["tenant_name"] = os.environ["SWIFT_TENANTNAME"]
        if "SWIFT_REGIONNAME" in os.environ:
            os_options.update({"region_name": os.environ["SWIFT_REGIONNAME"]})

        # formatting options for swiftclient.SwiftService
        for key in os_options.keys():
            svc_options[f"os_{key}"] = os_options[key]

        conn_kwargs["os_options"] = os_options

        # This folds the null prefix and all null parts, which means that:
        #  //MyContainer/ and //MyContainer are equivalent.
        #  //MyContainer//My/Prefix/ and //MyContainer/My/Prefix are equivalent.
        url_parts = [x for x in parsed_url.path.split("/") if x != ""]

        self.container = url_parts.pop(0)
        if url_parts:
            self.prefix = f"{'/'.join(url_parts)}/"
        else:
            self.prefix = ""

        policy = config.swift_storage_policy
        policy_header = "X-Storage-Policy"

        container_metadata = None
        try:
            log.Debug(f"Starting connection with arguments:'{conn_kwargs}'")
            self.conn = Connection(**conn_kwargs)
            container_metadata = self.conn.head_container(self.container)
        except ClientException as e:
            log.Debug(f"Connection failed: {e.__class__.__name__} {str(e)}")
            pass
        except Exception as e:
            raise errors.FatalError(
                f"Connection failed: {e.__class__.__name__} {str(e)}",
                log.ErrorCode.connection_failed,
            )

        if container_metadata is None:
            log.Info(f"Creating container {self.container}")
            try:
                headers = dict([[policy_header, policy]]) if policy else None
                self.conn.put_container(self.container, headers=headers)
            except Exception as e:
                raise errors.FatalError(
                    f"Container creation failed: {e.__class__.__name__} {str(e)}",
                    log.ErrorCode.connection_failed,
                )
        elif policy and container_metadata[policy_header.lower()] != policy:
            raise errors.FatalError(
                f"Container '{self.container}' exists but its storage policy is "
                f"'{container_metadata[policy_header.lower()]}' not '{policy}'."
            )
        else:
            log.Debug(f"Container already created: {container_metadata}")

        # checking service connection
        try:
            log.Debug(f"Starting  Swiftservice: '{svc_options}'")
            self.svc = SwiftService(options=svc_options)
            container_stat = self.svc.stat(self.container)
        except ClientException as e:
            raise errors.FatalError(
                f"Connection failed: {e.__class__.__name__} {str(e)}",
                log.ErrorCode.connection_failed,
            )
        log.Debug(f"Container stats: {container_stat}")

    def _error_code(self, operation, e):  # pylint: disable=unused-argument
        if isinstance(e, self.resp_exc):
            if e.http_status == 404:
                return log.ErrorCode.backend_not_found

    def _put(self, source_path, remote_filename):
        lp = os.fsdecode(source_path.name)
        if config.mp_segment_size > 0:
            st = os.stat(lp)
            # only upload using Dynamic Large Object if mpvolsize is triggered
            if st.st_size >= config.mp_segment_size:
                log.Debug("Uploading Dynamic Large Object")
                mp = self.svc.upload(
                    self.container,
                    [SwiftUploadObject(lp, object_name=self.prefix + os.fsdecode(remote_filename))],
                    options={"segment_size": config.mp_segment_size},
                )
                uploads = [a for a in mp if "container" not in a["action"]]
                for upload in uploads:
                    if not upload["success"]:
                        raise BackendException(upload["traceback"])
                return
        rp = self.prefix + os.fsdecode(remote_filename)
        log.Debug(f"Uploading '{lp}' to '{rp}' in remote container '{self.container}'")
        self.conn.put_object(
            container=self.container,
            obj=self.prefix + os.fsdecode(remote_filename),
            contents=open(lp, "rb"),
        )

    def _get(self, remote_filename, local_path):
        headers, body = self.conn.get_object(
            self.container,
            self.prefix + os.fsdecode(remote_filename),
            resp_chunk_size=1024,
        )
        with open(local_path.name, "wb") as f:
            for chunk in body:
                f.write(chunk)

    def _list(self):
        """
        List files in the remote Swift container.

        :returns: A list of filenames as byte strings.
        :rtype: list
        """
        headers, objs = self.conn.get_container(self.container, full_listing=True, path=self.prefix)
        # removes prefix from return values. should check for the prefix ?
        return [os.fsencode(o["name"][len(self.prefix) :]) for o in objs]

    def _delete(self, filename):
        """
        Delete a file from the Swift container.

        :param str filename: The name of the file to delete.
        """
        # use swiftservice to correctly delete all segments in case of multipart uploads
        deleted = [a for a in self.svc.delete(self.container, [self.prefix + os.fsdecode(filename)])]

    def _query(self, filename):
        """
        Get metadata for a file in the Swift container.

        :param str filename: The name of the file to query.
        :returns: A dictionary containing metadata (e.g., 'size').
        :rtype: dict
        """
        # use swiftservice to correctly report filesize in case of multipart uploads
        sobject = [a for a in self.svc.stat(self.container, [self.prefix + os.fsdecode(filename)])][0]
        sobj = {"size": int(sobject["headers"]["content-length"])}
        log.Debug(f"Objectquery: '{os.fsdecode(filename)}' has size {sobj['size']}.")
        return sobj


duplicity.backend.register_backend("swift", SwiftBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2012 Google Inc.
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Cache-wrapped functions for grp and pwd lookups.
"""

import grp
import pwd
from functools import lru_cache

limit = 512


@lru_cache(maxsize=limit)
def getgrgid(gid):
    """
    Return the group database entry for the given group ID.

    :param int gid: The group ID to look up.
    :returns: The group database entry.
    """
    return grp.getgrgid(gid)


@lru_cache(maxsize=limit)
def getgrnam(name):
    """
    Return the group database entry for the given group name.

    :param str name: The group name to look up.
    :returns: The group database entry or None if not found.
    """
    try:
        return grp.getgrnam(name)
    except KeyError:
        return None


@lru_cache(maxsize=limit)
def getpwnam(name):
    """
    Return the password database entry for the given user name.

    :param str name: The user name to look up.
    :returns: The password database entry or None if not found.
    """
    try:
        return pwd.getpwnam(name)
    except KeyError:
        return None


@lru_cache(maxsize=limit)
def getpwuid(uid):
    """
    Return the password database entry for the given user ID.

    :param int uid: The user ID to look up.
    :returns: The password database entry.
    """
    return pwd.getpwuid(uid)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Create and edit manifest for session contents
"""

import os
import re

from duplicity import config
from duplicity import errors
from duplicity import log
from duplicity import util


class ManifestError(Exception):
    """
    Exception raised when problem with manifest
    """

    pass


class Manifest(object):
    """
    List of volumes and information about each one
    """

    def __init__(self, fh=None):
        """
        Create a blank Manifest.

        :param DupPath fh: file object for the manifest.
        :returns: A new Manifest instance.
        :rtype: Manifest
        """
        self.corrupt_filelist = False
        self.hostname = None
        self.local_dirname = None
        self.volume_info_dict = {}  # dictionary vol numbers -> vol infos
        self.fh = fh
        self.files_changed = []

    def set_dirinfo(self):
        """
        Set information about directory from config, and write to manifest file.

        :returns: The manifest instance itself.
        :rtype: Manifest
        """
        self.hostname = config.hostname
        self.local_dirname = config.local_path.name
        if self.fh:
            if self.hostname:
                self.fh.write(b"Hostname %s\n" % self.hostname.encode())
            if self.local_dirname:
                self.fh.write(b"Localdir %s\n" % Quote(self.local_dirname))
        return self

    def check_dirinfo(self):
        """
        Check if directory information is the same as in config.

        Does not raise an error message if hostname or local_dirname
        are not available.

        :returns: None if dirinfo is the same.
        :rtype: str or None
        :raises FatalError: If the hostname or source path has changed.
        """
        if config.allow_source_mismatch:
            return

        # Check both hostname and fqdn (we used to write the fqdn into the
        # manifest, so we want to keep comparing against that)
        if self.hostname and self.hostname != config.hostname and self.hostname != config.fqdn:
            errmsg = _(
                "Fatal Error: Backup source host has changed.\n" "Current hostname: %s\n" "Previous hostname: %s"
            ) % (config.hostname, self.hostname)
            code = log.ErrorCode.hostname_mismatch
            code_extra = f"{util.escape(config.hostname)} {util.escape(self.hostname)}"

        elif self.local_dirname and self.local_dirname != config.local_path.name:
            errmsg = _(
                f"Fatal Error: Backup source directory has changed.\n"
                f"Current directory: {config.local_path.uc_name}\n"
                f"Previous directory: {os.fsdecode(self.local_dirname)}"
            )
            code = log.ErrorCode.source_path_mismatch
            code_extra = f"{util.escape(config.local_path.name)} {util.escape(self.local_dirname)}"
        else:
            return

        raise errors.FatalError(
            errmsg
            + "\n\n"
            + _(
                "Aborting because you may have accidentally tried to "
                "backup two different data sets to the same remote "
                "location, or using the same archive directory.  If "
                "this is not a mistake, use the "
                "--allow-source-mismatch switch to avoid seeing this "
                "message"
            ),
            code,
            code_extra,
        )

    def set_files_changed_info(self, files_changed):
        """
        Set the list of files that changed and write it to the manifest.
        """
        if files_changed:
            self.files_changed = files_changed

        if self.fh:
            self.fh.write(b"Filelist %d\n" % len(self.files_changed))
            for fileinfo in self.files_changed:
                self.fh.write(b"    %-7s  %s\n" % (fileinfo[1], Quote(fileinfo[0])))

    def add_volume_info(self, vi):
        """
        Add volume info to manifest and write it to the manifest file.

        :param VolumeInfo vi: The volume info to add.
        """
        vol_num = vi.volume_number
        self.volume_info_dict[vol_num] = vi
        if self.fh:
            self.fh.write(vi.to_string() + b"\n")

    def del_volume_info(self, vol_num):
        """
        Remove volume ``vol_num`` from the manifest.

        :param int vol_num: The volume number to delete.
        :raises ManifestError: If the volume is not present in the manifest.
        """
        try:
            del self.volume_info_dict[vol_num]
        except Exception:
            raise ManifestError(f"Volume {int(vol_num)} not present in manifest")

    def to_string(self):
        """
        Return the string version of the manifest.

        Just concatenates the volume info strings.

        :returns: The manifest in string form.
        :rtype: bytes
        """
        result = b""
        if self.hostname:
            result += b"Hostname %s\n" % self.hostname.encode()
        if self.local_dirname:
            result += b"Localdir %s\n" % Quote(self.local_dirname)

        result += b"Filelist %d\n" % len(self.files_changed)
        for fileinfo in self.files_changed:
            result += b"    %-7s  %s\n" % (fileinfo[1], Quote(fileinfo[0]))

        vol_num_list = sorted(self.volume_info_dict.keys())

        def vol_num_to_string(vol_num):
            return self.volume_info_dict[vol_num].to_string()

        result = b"%s%s\n" % (result, b"\n".join(map(vol_num_to_string, vol_num_list)))
        return result

    __str__ = to_string

    def from_string(self, s):
        """
        Initialize the manifest from a string.

        :param bytes s: The byte string to parse.
        :returns: The manifest instance.
        :rtype: Manifest
        """

        def get_field(fieldname):
            """
            Return the value of a field by parsing s, or None if no field
            """
            if not isinstance(fieldname, bytes):
                fieldname = fieldname.encode()
            m = re.search(b"(^|\\n)%s\\s(.*?)\n" % fieldname, s, re.I)
            if not m:
                return None
            else:
                return Unquote(m.group(2))

        self.hostname = get_field("hostname")
        if self.hostname is not None:
            self.hostname = self.hostname.decode()
        self.local_dirname = get_field("localdir")

        highest_vol = 0
        latest_vol = 0
        vi_regexp = re.compile(b"(?:^|\\n)(volume\\s.*(?:\\n.*)*?)(?=\\nvolume\\s|$)", re.I)
        vi_iterator = vi_regexp.finditer(s)
        for match in vi_iterator:
            vi = VolumeInfo().from_string(match.group(1))
            self.add_volume_info(vi)
            latest_vol = vi.volume_number
            highest_vol = max(highest_vol, latest_vol)
            log.Debug(_("Found manifest volume %s") % latest_vol)
        # If we restarted after losing some remote volumes, the highest volume
        # seen may be higher than the last volume recorded.  That is, the
        # manifest could contain "vol1, vol2, vol3, vol2."  If so, we don't
        # want to keep vol3's info.
        for i in range(latest_vol + 1, highest_vol + 1):
            self.del_volume_info(i)
        log.Info(_("Found %s volumes in manifest") % latest_vol)

        # Get file changed list - not needed if --file-changed and
        # --show-changes-in-set are not present
        filecount = 0
        if config.file_changed is not None or config.show_changes_in_set is not None:
            filelist_regexp = re.compile(b"(^|\\n)filelist\\s([0-9]+)\\n(.*?)(\\nvolume\\s|$)", re.I | re.S)
            match = filelist_regexp.search(s)
            if match:
                filecount = int(match.group(2))
            if filecount > 0:

                def parse_fileinfo(line):
                    fileinfo = line.strip().split()
                    return fileinfo[0], b"".join(fileinfo[1:])

                self.files_changed = list(map(parse_fileinfo, match.group(3).split(b"\n")))

            if filecount != len(self.files_changed):
                log.Error(
                    _(
                        f"Manifest file '{self.fh.base if self.fh else ''}' is corrupt: "
                        f"File count says {int(filecount)}, File list contains {len(self.files_changed)}"
                    )
                )
                self.corrupt_filelist = True

        return self

    def get_files_changed(self):
        """
        Return the list of files that changed.
        """
        return self.files_changed

    def __eq__(self, other):
        """
        Two manifests are equal if they contain the same volume infos
        """
        vi_list1 = sorted(self.volume_info_dict.keys())
        vi_list2 = sorted(other.volume_info_dict.keys())

        if vi_list1 != vi_list2:
            log.Notice(_("Manifests not equal because different volume numbers"))
            return False

        for i in range(len(vi_list1)):
            if not vi_list1[i] == vi_list2[i]:
                log.Notice(_("Manifests not equal because volume lists differ"))
                return False

        if self.hostname != other.hostname or self.local_dirname != other.local_dirname:
            log.Notice(_("Manifests not equal because hosts or directories differ"))
            return False

        return True

    def __ne__(self, other):
        """
        Defines !=.  Not doing this always leads to annoying bugs...
        """
        return not self.__eq__(other)

    def write_to_path(self, path):
        """
        Write the string version of the manifest to the given path.

        :param Path path: The path to write to.
        """
        assert not path.exists(), "Destination path already exists; refusing to overwrite manifest"
        fout = path.open("wb")
        fout.write(self.to_string())
        assert not fout.close(), "fout failed to close"
        path.setdata()

    def get_containing_volumes(self, index_prefix):
        """
        Return list of volume numbers that may contain index_prefix
        """
        if len(index_prefix) == 1 and isinstance(index_prefix[0], str):
            index_prefix = (index_prefix[0].encode(),)
        return [
            vol_num
            for vol_num in list(self.volume_info_dict.keys())
            if self.volume_info_dict[vol_num].contains(index_prefix)
        ]


class VolumeInfoError(Exception):
    """
    Raised when there is a problem initializing a VolumeInfo from string
    """

    pass


class VolumeInfo(object):
    """
    Information about a single volume
    """

    def __init__(self):
        """
        VolumeInfo initializer
        """
        self.volume_number = None
        self.start_index = None
        self.start_block = None
        self.end_index = None
        self.end_block = None
        self.hashes = {}

    def set_info(self, vol_number, start_index, start_block, end_index, end_block):
        """
        Set essential VolumeInfo information, return self

        Call with starting and ending paths stored in the volume.  If
        a multivol diff gets split between volumes, count it as being
        part of both volumes.
        """
        self.volume_number = vol_number
        self.start_index = start_index
        self.start_block = start_block
        self.end_index = end_index
        self.end_block = end_block

        return self

    def set_hash(self, hash_name, data):
        """
        Set the value of hash hash_name (e.g. "MD5") to data
        """
        if isinstance(hash_name, bytes):
            hash_name = hash_name.decode()
        if isinstance(data, bytes):
            data = data.decode()
        self.hashes[hash_name] = data

    def get_best_hash(self):
        """
        Return pair (hash_type, hash_data)

        SHA1 is the best hash, and MD5 is the second best hash.  None
        is returned if no hash is available.
        """
        if not self.hashes:
            return None
        try:
            return "SHA1", self.hashes["SHA1"]
        except KeyError:
            pass
        try:
            return "MD5", self.hashes["MD5"]
        except KeyError:
            pass
        return list(self.hashes.items())[0]

    def to_string(self):
        """
        Return nicely formatted string reporting all information
        """

        def index_to_string(index):
            """
            Return printable version of index without any whitespace
            """
            if index:
                s = b"/".join(index)
                return Quote(s)
            else:
                return b"."

        def bfmt(x):
            if x is None:
                return b" "
            return str(x).encode()

        slist = [b"Volume %d:" % self.volume_number]
        whitespace = b"    "
        slist.append(
            b"%sStartingPath   %s %s" % (whitespace, index_to_string(self.start_index), bfmt(self.start_block))
        )
        slist.append(b"%sEndingPath     %s %s" % (whitespace, index_to_string(self.end_index), bfmt(self.end_block)))
        for key in self.hashes:
            slist.append(b"%sHash %s %s" % (whitespace, key.encode(), self.hashes[key].encode()))
        return b"\n".join(slist)

    __str__ = to_string

    def from_string(self, s):
        """
        Initialize self from string s as created by to_string
        """

        def string_to_index(s):
            """
            Return tuple index from string
            """
            s = Unquote(s)
            if s == b".":
                return ()
            return tuple(s.split(b"/"))

        linelist = s.strip().split(b"\n")

        # Set volume number
        m = re.search(b"^Volume ([0-9]+):", linelist[0], re.I)
        if not m:
            raise VolumeInfoError(f"Bad first line '{linelist[0]}'")
        self.volume_number = int(m.group(1))

        # Set other fields
        for line in linelist[1:]:
            if not line:
                continue
            line_split = line.strip().split()
            field_name = line_split[0].lower()
            other_fields = line_split[1:]
            if field_name == b"Volume":
                log.Warn(_("WARNING. found extra Volume identifier"))
                break
            elif field_name == b"startingpath":
                self.start_index = string_to_index(other_fields[0])
                if len(other_fields) > 1:
                    self.start_block = int(other_fields[1])
                else:
                    self.start_block = None
            elif field_name == b"endingpath":
                self.end_index = string_to_index(other_fields[0])
                if len(other_fields) > 1:
                    self.end_block = int(other_fields[1])
                else:
                    self.end_block = None
            elif field_name == b"hash":
                self.set_hash(other_fields[0], other_fields[1])

        if self.start_index is None or self.end_index is None:
            raise VolumeInfoError("Start or end index not set")
        return self

    def __eq__(self, other):
        """
        Used in test suite
        """
        if not isinstance(other, VolumeInfo):
            log.Notice(_("Other is not VolumeInfo"))
            return None
        if self.volume_number != other.volume_number:
            log.Notice(_("Volume numbers don't match"))
            return None
        if self.start_index != other.start_index:
            log.Notice(_("start_indicies don't match"))
            return None
        if self.end_index != other.end_index:
            log.Notice(_("end_index don't match"))
            return None
        hash_list1 = sorted(self.hashes.items())
        hash_list2 = sorted(other.hashes.items())
        if hash_list1 != hash_list2:
            log.Notice(_("Hashes don't match"))
            return None
        return 1

    def __ne__(self, other):
        """
        Defines !=
        """
        return not self.__eq__(other)

    def contains(self, index_prefix, recursive=True):
        """
        Return true if volume might contain the given index.

        If recursive is true, then return true if any index starting
        with index_prefix could be contained.  Otherwise, just check
        if index_prefix itself is between the starting and ending
        indices.

        :param tuple index_prefix: The path index to check.
        :param bool recursive: Whether to perform a recursive check.
        :returns: True if the index (or its children) might be in this volume.
        :rtype: bool
        """
        if recursive:
            return self.start_index[: len(index_prefix)] <= index_prefix <= self.end_index
        else:
            return self.start_index <= index_prefix <= self.end_index


nonnormal_char_re = re.compile(b"(\\s|[\\\\\"'])")


def Quote(s):
    """
    Return quoted version of s safe to put in a manifest or volume info.

    :param bytes s: The byte string to quote.
    :returns: The quoted byte string.
    :rtype: bytes
    """
    if not nonnormal_char_re.search(s):
        return s  # no quoting necessary
    slist = []
    for i in range(0, len(s)):
        char = s[i : i + 1]
        if nonnormal_char_re.search(char):
            slist.append(b"\\x%02x" % ord(char))
        else:
            slist.append(char)
    return b'"%s"' % b"".join(slist)


def maybe_chr(ch):
    """
    Return the character for the given ordinal.
    """
    return chr(ch)


def Unquote(quoted_string):
    """
    Return the original string from a quoted string produced by :func:`Quote`.

    :param bytes quoted_string: The quoted byte string to unquote.
    :returns: The original byte string.
    :rtype: bytes
    """
    if not maybe_chr(quoted_string[0]) == '"' or maybe_chr(quoted_string[0]) == "'":
        return quoted_string
    assert quoted_string[0] == quoted_string[-1], "Quoted string must start and end with the same quote character"
    return_list = []
    i = 1  # skip initial char
    while i < len(quoted_string) - 1:
        char = quoted_string[i : i + 1]
        if char == b"\\":
            # quoted section
            assert maybe_chr(quoted_string[i + 1]) == "x", "Escape sequence must be of the form \\xNN (hex)"
            return_list.append(int(quoted_string[i + 2 : i + 4].decode(), 16).to_bytes(1, byteorder="big"))
            i += 4
        else:
            return_list.append(char)
            i += 1
    return b"".join(return_list)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2022 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Data for parse command line, check for consistency, and set config
"""

from dataclasses import dataclass

from duplicity import __reldate__
from duplicity import __version__
from duplicity.cli_util import *


@dataclass(order=True)
class DuplicityCommands:
    """
    Duplicity commands and their expected positional arguments.

    .. note::
       ``cli_util`` must contain a function named ``check_*`` for each positional
       argument (e.g., ``check_source_path()``).
    """

    backup = ["source_path", "target_url"]
    cleanup = ["target_url"]
    collection_status = ["target_url"]
    full = ["source_path", "target_url"]
    incremental = ["source_path", "target_url"]
    list_current_files = ["target_url"]
    remove_all_but_n_full = ["count", "target_url"]
    remove_all_inc_of_but_n_full = ["count", "target_url"]
    remove_older_than = ["remove_time", "target_url"]
    restore = ["source_url", "target_path"]
    verify = ["source_url", "target_path"]


@dataclass(order=True)
class CommandAliases:
    """
    Duplicity commands and their aliases.
    """

    backup = ["bu"]
    cleanup = ["cl"]
    collection_status = ["st"]
    full = ["fb"]
    incremental = ["incr", "inc", "ib"]
    list_current_files = ["ls"]
    remove_all_but_n_full = ["ra"]
    remove_all_inc_of_but_n_full = ["ri"]
    remove_older_than = ["ro"]
    restore = ["rb"]
    verify = ["vb"]


all_commands = set()
for var, aliases in CommandAliases.__dict__.items():
    if var.startswith("__"):
        continue
    all_commands.add(var2cmd(var))
    for alias in aliases:
        all_commands.add(alias)


command_args_expected = dict()
for var, aliases in CommandAliases.__dict__.items():
    if var.startswith("__"):
        continue
    cmd = var2cmd(var)
    expect = len(DuplicityCommands.__dict__[var])
    command_args_expected[cmd] = expect
    for alias in aliases:
        command_args_expected[alias] = expect


"""
Option kwargs for add_argument
"""
OptionKwargs = dict(
    allow_source_mismatch=dict(
        action="store_true",
        help="Allow different source directories",
        default=dflt(config.allow_source_mismatch),
    ),
    archive_dir=dict(
        metavar=_("path"),
        type=check_file,
        help="Path to store metadata archives",
        default=dflt(config.archive_dir),
    ),
    asynchronous_upload=dict(
        action=WarnAsyncStoreConstAction,
        const=1,
        dest="async_concurrency",
        help="Discontiniued, use --concurrency",
        default=dflt(config.async_concurrency),
    ),
    azure_blob_tier=dict(
        metavar=_("Hot|Cool|Cold|Archive"),
        help="Standard storage tier used for storing backup files (Hot|Cool|Cold|Archive)",
        default=dflt(config.azure_blob_tier),
    ),
    azure_max_connections=dict(
        metavar=_("number"),
        type=int,
        help="Number of maximum parallel connections to use when the blob size exceeds 64MB",
        default=dflt(config.azure_max_connections),
    ),
    azure_max_block_size=dict(
        metavar=_("number"),
        type=int,
        help="Number for the block size to upload a blob if the length is unknown\n"
        "or is larger than the value set by --azure-max-single-put-size\n"
        "The maximum block size the service supports is 100MiB.",
        default=dflt(config.azure_max_block_size),
    ),
    azure_max_single_put_size=dict(
        metavar=_("number"),
        type=int,
        help="Largest supported upload size where the Azure library makes only one put call.\n"
        "Used to upload a single block if the content length is known and is less than this",
        default=dflt(config.azure_max_single_put_size),
    ),
    b2_hide_files=dict(
        action="store_true",
        help="Whether the B2 backend hides files instead of deleting them",
        default=dflt(config.b2_hide_files),
    ),
    backend_retry_delay=dict(
        metavar=_("seconds"),
        type=int,
        help="Delay time before next try after a failure of a backend operation",
        default=dflt(config.backend_retry_delay),
    ),
    cf_backend=dict(
        metavar="pyrax|cloudfiles",
        help="Allow the user to switch cloudfiles backend",
        default=dflt(config.cf_backend),
    ),
    compare_data=dict(
        action="store_true",
        help="Compare data on verify not only signatures",
        default=dflt(config.compare_data),
    ),
    concurrency=dict(
        metavar=_("number"),
        type=int,
        help="Number n of concurrent backend processes to get better performance.\n"
        "Allow parallel creation of new volume and transfer. Some backends may profit more than one transfer, too.\n"
        "Typically a value of 1-4 is a good starting point. The impact totally depends on the backend,\n"
        "network and other condition. This will also increase the temporary local disk usage\n"
        "from 1 to n + 1 volumes.",
        default=dflt(config.concurrency),
    ),
    config_dir=dict(
        metavar=_("path"),
        type=check_file,
        help="Path to store configuration files",
        default=dflt(config.config_dir),
    ),
    copy_blocksize=dict(
        metavar=_("number"),
        type=set_kilos,
        help="Blocksize to use in copy operations in kilobytes.",
        default=dflt(config.copy_blocksize),
    ),
    copy_links=dict(
        action="store_true",
        help="Copy contents of symlinks instead of linking",
        default=dflt(config.copy_links),
    ),
    dry_run=dict(
        action="store_true",
        help="Perform dry-run with no writes",
        default=dflt(config.dry_run),
    ),
    encrypt_key=dict(
        metavar=_("gpg-key-id"),
        type=set_encrypt_key,
        help="GNUpg key for encryption/decryption",
        default=dflt(None),
    ),
    encrypt_sign_key=dict(
        metavar=_("gpg-key-id"),
        type=set_encrypt_sign_key,
        help="GNUpg key for encryption/decryption and signing",
        default=dflt(None),
    ),
    encrypt_secret_keyring=dict(
        metavar=_("path"),
        help="Path to secret GNUpg keyring",
        default=dflt(None),
    ),
    exclude=dict(
        metavar=_("shell_pattern"),
        action=AddSelectionAction,
        help="Exclude globbing pattern",
        default=dflt(None),
    ),
    exclude_device_files=dict(
        nargs=0,
        action=AddSelectionAction,
        help="Exclude device files",
        default=dflt(False),
    ),
    exclude_filelist=dict(
        metavar=_("filename"),
        action=AddFilelistAction,
        help="File with list of file patterns to exclude",
        default=dflt(None),
    ),
    exclude_if_present=dict(
        metavar=_("filename"),
        action=AddSelectionAction,
        help="Exclude directory if this file is present",
        default=dflt(None),
    ),
    exclude_older_than=dict(
        metavar=_("time"),
        type=check_time,
        action=AddSelectionAction,
        help="Exclude files older than time",
        default=dflt(None),
    ),
    exclude_other_filesystems=dict(
        nargs=0,
        action=AddSelectionAction,
        help="Exclude other filesystems from backup",
        default=dflt(False),
    ),
    exclude_regexp=dict(
        metavar=_("regex"),
        action=AddSelectionAction,
        help="Exclude based on regex pattern",
        default=dflt(None),
    ),
    file_changed=dict(
        metavar=_("path"),
        type=check_file,
        help="Whether to collect only the file status, not the whole root",
        default=dflt(None),
    ),
    file_prefix=dict(
        metavar="string",
        type=make_bytes,
        help="String prefix for all duplicity files",
        default=dflt(config.file_prefix),
    ),
    file_prefix_archive=dict(
        metavar="string",
        type=make_bytes,
        help="String prefix for duplicity difftar files",
        default=dflt(config.file_prefix_archive),
    ),
    file_prefix_manifest=dict(
        metavar="string",
        type=make_bytes,
        help="String prefix for duplicity manifest files",
        default=dflt(config.file_prefix_manifest),
    ),
    file_prefix_signature=dict(
        metavar="string",
        type=make_bytes,
        help="String prefix for duplicity signature files",
        default=dflt(config.file_prefix_signature),
    ),
    file_prefix_jsonstat=dict(
        metavar="string",
        type=make_bytes,
        help="String prefix for duplicity jsonstat files",
        default=dflt(config.file_prefix_jsonstat),
    ),
    files_from=dict(
        metavar=_("filename"),
        type=check_file,
        action=AddFilelistAction,
        help="Defines the backup source as a sub-set of the source folder",
        default=dflt(None),
    ),
    filter_globbing=dict(
        nargs=0,
        action=AddSelectionAction,
        help="File selection mode switch, changes the interpretation of any subsequent\n"
        "--exclude* or --include* options to shell globbing.",
        default=dflt(None),
    ),
    filter_ignorecase=dict(
        nargs=0,
        action=AddSelectionAction,
        help="File selection mode switch, changes the interpretation of any subsequent\n"
        "--exclude* or --include* options to case-insensitive matching.",
        default=dflt(None),
    ),
    filter_literal=dict(
        nargs=0,
        action=AddSelectionAction,
        help="File selection mode switch, changes the interpretation of any subsequent\n"
        "--exclude* or --include* options to literal strings.",
        default=dflt(None),
    ),
    filter_regexp=dict(
        nargs=0,
        action=AddSelectionAction,
        help="File selection mode switch, changes the interpretation of any subsequent\n"
        "--exclude* or --include* options to regular expressions.",
        default=dflt(None),
    ),
    filter_strictcase=dict(
        nargs=0,
        action=AddSelectionAction,
        help="File selection mode switch, changes the interpretation of any subsequent\n"
        "--exclude* or --include* options to case-sensitive matching.",
        default=dflt(None),
    ),
    force=dict(
        action="store_true",
        help="Force duplicity to actually delete during cleanup",
        default=dflt(config.force),
    ),
    ftp_passive=dict(
        action="store_const",
        const="passive",
        dest="ftp_connection",
        help="Tell FTP to use passive mode",
        default=dflt(config.ftp_connection),
    ),
    ftp_regular=dict(
        action="store_const",
        const="regular",
        dest="ftp_connection",
        help="Tell FTP to use regular mode",
        default=dflt(config.ftp_connection),
    ),
    full_if_older_than=dict(
        metavar=_("time"),
        type=check_time_delta,
        help="Perform full backup if last full is older than 'time'",
        default=dflt(config.full_if_older_than),
    ),
    full_if_n_inc=dict(
        metavar=_("number"),
        type=int,
        help="Perform full backup if latest chain has N or more incremental backups",
        default=dflt(config.full_if_n_inc),
    ),
    gpg_binary=dict(
        metavar=_("path"),
        type=check_file,
        help="Path to GNUpg executable file",
        default=dflt(config.gpg_binary),
    ),
    gpg_options=dict(
        metavar=_("options"),
        action=SplitOptionsAction,
        help=(
            "Verbatim gpg options.  May be supplied multiple times.\n"
            "NOTE: --gpg-options and value should be bound with an '=' as in\n"
            "      --gpg-options='--some-option=value'\n"
            "See man page at ARGPARSE PROBLEM."
        ),
        default=dflt(config.gpg_options),
    ),
    use_gpgsm=dict(
        action="store_true",
        dest="use_gpgsm",
        help="Use gpgsm tool, which is very similar gpg but using PKCS#7 file format.",
        default=dflt(config.use_gpgsm),
    ),
    hidden_encrypt_key=dict(
        metavar=_("gpg-key-id"),
        type=set_hidden_encrypt_key,
        help="Hidden GNUpg encryption key",
        default=dflt(None),
    ),
    idr_fakeroot=dict(
        metavar=_("path"),
        type=check_file,
        help="Fake root for idrive backend",
        default=dflt(config.idr_fakeroot),
    ),
    ignore_errors=dict(
        nargs=0,
        action=IgnoreErrorsAction,
        help="Ignore most errors during processing",
        default=dflt(config.ignore_errors),
    ),
    imap_full_address=dict(
        action="store_true",
        help="Whether to use the full email address as the user name",
        default=dflt(config.imap_full_address),
    ),
    imap_mailbox=dict(
        metavar=_("imap_mailbox"),
        help="Name of the imap folder to store backups",
        default=dflt(config.imap_mailbox),
    ),
    include=dict(
        metavar=_("shell_pattern"),
        action=AddSelectionAction,
        help="Include globbing pattern",
        default=dflt(None),
    ),
    include_filelist=dict(
        metavar=_("filename"),
        action=AddFilelistAction,
        help="File with list of file patterns to include",
        default=dflt(None),
    ),
    include_regexp=dict(
        metavar=_("regex"),
        action=AddSelectionAction,
        help="Include based on regex pattern",
        default=dflt(None),
    ),
    jsonstat=dict(
        action="store_true",
        help=_(
            "If set, an extra file with statistics in json format will be stored with each backup. "
            "It includes the statistics printed out on stdout by default. "
            "With 'collection-status' it returs these stats."
        ),
        default=dflt(config.jsonstat),
    ),
    # log_fd is directly applied in set_log_fd(), not saved in config
    log_fd=dict(
        metavar=_("file_descriptor"),
        dest="",
        type=set_log_fd,
        help="File descriptor to be used for logging",
    ),
    # log_file is directly applied in set_log_file(), not saved in config
    log_file=dict(
        metavar=_("log_filename"),
        dest="",
        type=set_log_file,
        help="Logging filename to use",
    ),
    # log_timestamp is directly applied in set_log_timestamp(), not saved in config
    log_timestamp=dict(
        nargs=0,
        action=SetLogTimestampAction,
        help="Whether to include timestamp and level in log",
        default=dflt(False),
    ),
    max_blocksize=dict(
        metavar=_("number"),
        type=round512,
        help=(
            "The integer square root of file length is used as the librsync block size\n"
            "up to this maximum block size value in bytes if supplied.\n"
            "Block size is rounded up to the nearest 512 byte boundary."
        ),
        default=dflt(config.max_blocksize),
    ),
    metadata_sync_mode=dict(
        choices=("full", "partial"),
        dest="metadata_sync_mode",
        help="Only sync required metadata not all",
        default=dflt(config.metadata_sync_mode),
    ),
    mf_purge=dict(
        action="store_true",
        help="Option for mediafire to purge files on delete instead of sending to trash",
        default=dflt(config.mf_purge),
    ),
    mp_segment_size=dict(
        metavar=_("number"),
        type=set_megs,
        help="Swift backend segment size in megabytes",
        default=dflt(config.mp_segment_size),
    ),
    name=dict(
        metavar=_("backup name"),
        dest="backup_name",
        help="Custom backup name instead of hash",
        default=dflt(config.backup_name),
    ),
    no_check_remote=dict(
        action="store_false",
        dest="check_remote",
        help="If supplied do not check remote metadata",
        default=dflt(config.check_remote),
    ),
    no_compression=dict(
        action="store_false",
        dest="compression",
        help="If supplied do not perform compression",
        default=dflt(config.compression),
    ),
    no_encryption=dict(
        action="store_false",
        dest="encryption",
        help="If supplied do not perform encryption",
        default=dflt(config.encryption),
    ),
    no_files_changed=dict(
        action="store_false",
        dest="files_changed",
        help="If supplied do not collect the files_changed list",
        default=dflt(config.files_changed),
    ),
    no_restore_ownership=dict(
        action="store_false",
        dest="restore_ownership",
        help="If supplied do not restore uid/gid when finished",
        default=dflt(config.restore_ownership),
    ),
    no_print_statistics=dict(
        action="store_false",
        dest="print_statistics",
        help="If supplied do not print statistics",
        default=dflt(config.print_statistics),
    ),
    null_separator=dict(
        action="store_true",
        help="Whether to split on null instead of newline",
        default=dflt(config.null_separator),
    ),
    num_retries=dict(
        metavar=_("number"),
        type=int,
        help="Number of retries on network operations",
        default=dflt(config.num_retries),
    ),
    numeric_owner=dict(
        action="store_true",
        help="Keeps number from tar file. Like same option in GNU tar.",
        default=dflt(config.numeric_owner),
    ),
    par2_options=dict(
        metavar=_("options"),
        action=SplitOptionsAction,
        help=(
            "Verbatim par2 options.  May be supplied multiple times.\n"
            "NOTE: --par2-options and value should be bound with an '=' as in\n"
            "      --par2-options='--some-option=value'\n"
            "See man page at ARGPARSE PROBLEM."
        ),
        default=dflt(config.par2_options),
    ),
    par2_redundancy=dict(
        metavar=_("number"),
        type=int,
        help="Level of Redundancy in percent for Par2 files",
        default=dflt(config.par2_redundancy),
    ),
    par2_volumes=dict(
        metavar=_("number"),
        type=int,
        help="Number of par2 volumes",
        default=dflt(config.par2_volumes),
    ),
    path_to_restore=dict(
        metavar=_("path"),
        type=check_file,
        dest="restore_path",
        help="File or directory path to restore",
        default=dflt(config.restore_path),
    ),
    progress=dict(
        action="store_true",
        help="Display progress for the full and incremental backup operations",
        default=dflt(config.progress),
    ),
    progress_rate=dict(
        metavar=_("number"),
        type=int,
        help="Used to control the progress option update rate in seconds",
        default=dflt(config.progress_rate),
    ),
    rename=dict(
        metavar="from to",
        action=AddRenameAction,
        nargs=2,
        help="Rename files during restore",
        default=dflt(config.rename),
    ),
    restore_time=dict(
        metavar=_("time"),
        type=check_time,
        help="Restores will try to bring back the state as of the following time",
        default=dflt(config.restore_time),
    ),
    rsync_options=dict(
        metavar=_("options"),
        action=SplitOptionsAction,
        help=(
            "Verbatim rsync options.  May be supplied multiple times.\n"
            "NOTE: --rsync-options and value should be bound with an '=' as in\n"
            "      --rsync-options='--some-option=value'\n"
            "See man page at ARGPARSE PROBLEM."
        ),
        default=dflt(config.rsync_options),
    ),
    s3_endpoint_url=dict(
        metavar=_("s3_endpoint_url"),
        action="store",
        help="Specity S3 endpoint",
        default=dflt(config.s3_endpoint_url),
    ),
    s3_unencrypted_connection=dict(
        action="store_true",
        help="Whether to use plain HTTP (without SSL) to send data to S3",
        default=dflt(config.s3_unencrypted_connection),
    ),
    s3_use_deep_archive=dict(
        action="store_true",
        help="Whether to use S3 Glacier Deep Archive Storage",
        default=dflt(config.s3_use_deep_archive),
    ),
    s3_use_glacier=dict(
        action="store_true",
        help="Whether to use S3 Glacier Storage",
        default=dflt(config.s3_use_glacier),
    ),
    s3_use_glacier_ir=dict(
        action="store_true",
        help="Whether to use S3 Glacier IR Storage",
        default=dflt(config.s3_use_glacier_ir),
    ),
    s3_use_ia=dict(
        action="store_true",
        help="Whether to use S3 Infrequent Access Storage",
        default=dflt(config.s3_use_ia),
    ),
    s3_use_onezone_ia=dict(
        action="store_true",
        help="Whether to use S3 One Zone Infrequent Access Storage",
        default=dflt(config.s3_use_onezone_ia),
    ),
    s3_use_rrs=dict(
        action="store_true",
        help="Whether to use S3 Reduced Redundancy Storage",
        default=dflt(config.s3_use_rrs),
    ),
    s3_multipart_chunk_size=dict(
        metavar=_("number"),
        type=set_megs,
        help="Chunk size used for S3 multipart uploads.The number of parallel uploads to\n"
        "S3 be given by chunk size / volume size. Use this to maximize the use of\n"
        "your bandwidth",
        default=dflt(config.s3_multipart_chunk_size),
    ),
    s3_multipart_max_procs=dict(
        metavar=_("number"),
        type=int,
        help="Number of processes to set the Processor Pool to when uploading multipart\n"
        "uploads to S3. Use this to control the maximum simultaneous uploads to S3",
        default=dflt(config.s3_multipart_max_procs),
    ),
    s3_use_server_side_encryption=dict(
        action="store_true",
        dest="s3_use_sse",
        help="Allow use of server side encryption",
        default=dflt(config.s3_use_sse),
    ),
    s3_use_server_side_kms_encryption=dict(
        action="store_true",
        dest="s3_use_sse_kms",
        help="Allow use of server side KMS encryption",
        default=dflt(config.s3_use_sse_kms),
    ),
    s3_kms_key_id=dict(
        metavar=_("s3_kms_key_id"),
        action="store",
        help="S3 KMS encryption key id",
        default=dflt(config.s3_kms_key_id),
    ),
    s3_kms_grant=dict(
        metavar=_("s3_kms_grant"),
        action="store",
        help="S3 KMS grant value",
        default=dflt(config.s3_kms_grant),
    ),
    s3_region_name=dict(
        metavar=_("s3_region_name"),
        action="store",
        help="Specity S3 region name",
        default=dflt(config.s3_region_name),
    ),
    swift_storage_policy=dict(
        metavar=_("policy"),
        help="Option to specify a Swift container storage policy.",
        default=dflt(config.swift_storage_policy),
    ),
    scp_command=dict(
        metavar=_("command"),
        help="SCP command to use (ssh pexpect backend)",
        default=dflt(config.scp_command),
    ),
    sftp_command=dict(
        metavar=_("command"),
        help="SFTP command to use (ssh pexpect backend)",
        default=dflt(config.sftp_command),
    ),
    show_changes_in_set=dict(
        metavar=_("number"),
        type=int,
        help="Show file changes (new, deleted, changed) in the specified backup\n"
        "set (0 specifies latest, 1 specifies next latest, etc.)",
        default=dflt(config.show_changes_in_set),
    ),
    sign_key=dict(
        metavar=_("gpg-key-id"),
        type=set_sign_key,
        help="Sign key for encryption/decryption",
        default=dflt(None),
    ),
    skip_if_no_change=dict(
        action="store_true",
        help="Skip incremental backup if nothing changed.",
        default=dflt(config.skip_if_no_change),
    ),
    ssh_askpass=dict(
        action="store_true",
        help="Ask the user for the SSH password. Not for batch usage",
        default=dflt(config.ssh_askpass),
    ),
    ssh_options=dict(
        metavar=_("options"),
        action=SplitOptionsAction,
        help=(
            "Verbatim ssh options.  May be supplied multiple times.\n"
            "NOTE: --ssh-options and value should be bound with an '=' as in\n"
            "      --ssh-options='--some-option=value'\n"
            "See man page at ARGPARSE PROBLEM."
        ),
        default=dflt(config.ssh_options),
    ),
    ssl_cacert_file=dict(
        metavar="file",
        help=_("pem formatted bundle of certificate authorities"),
        default=dflt(config.ssl_cacert_file),
    ),
    ssl_cacert_path=dict(
        metavar="path",
        help=_("path to a folder with certificate authority files"),
        default=dflt(config.ssl_cacert_path),
    ),
    ssl_no_check_certificate=dict(
        action="store_true",
        help="Set to not validate SSL certificates",
        default=dflt(config.ssl_no_check_certificate),
    ),
    tempdir=dict(
        metavar=_("path"),
        type=check_file,
        dest="temproot",
        help="Working directory for temp files",
        default=dflt(config.temproot),
    ),
    timeout=dict(
        metavar=_("seconds"),
        type=check_timeout,
        help="Network timeout in seconds",
        default=dflt(config.timeout),
    ),
    time_separator=dict(
        metavar=_("char"),
        type=check_char,
        help="Character used like the ':' in time strings like=2002-08-06T04:22:00-07:00\n"
        "Note=Not valid for backup commands full and inc, only to read older backups",
        default=dflt(config.time_separator),
    ),
    use_agent=dict(
        action="store_true",
        help="Whether to specify --use-agent in GnuPG options",
        default=dflt(config.use_agent),
    ),
    # verbosity is set directly via check_verbosity(), not saved in config
    verbosity=dict(
        metavar=_("verb"),
        dest="",
        type=check_verbosity,
        help="Logging verbosity=[0-9] or [e, w, n, i, d] or [error, warning, notice, info, debug]",
    ),
    version=dict(
        action="version",
        version=f"duplicity {__version__} {__reldate__}",
        help="Display version and exit",
    ),
    volsize=dict(
        metavar=_("number"),
        type=set_megs,
        help="Volume size to use in MiB",
        default=dflt(config.volsize),
    ),
    webdav_headers=dict(
        metavar="string",
        help=_("extra headers for Webdav, like 'Cookie,name=value'"),
        default=dflt(config.webdav_headers),
    ),
    # TESTING ONLY - do not use in production
    current_time=dict(
        metavar="time",
        type=int,
        help=argparse.SUPPRESS,
    ),
    pydevd=dict(
        # activate remote debugging
        action="store_true",
        help=argparse.SUPPRESS,
    ),
    skip_volume=dict(
        metavar="volume",
        type=int,
        help=argparse.SUPPRESS,
    ),
)


@dataclass(order=True)
class OptionAliases:
    """
    Aliases for command line options.
    """

    path_to_restore = ["-r"]
    restore_time = ["-t", "--time"]
    verbosity = ["-v"]
    version = ["-V"]


logging_options = {
    "--log-fd",
    "--log-file",
    "--log-timestamp",
    "--verbosity",
    "--version",
}

backup_only_options = {
    "--allow-source-mismatch",
    "--asynchronous-upload",
    "--dry-run",
    "--volsize",
}

selection_only_options = {
    "--exclude",
    "--exclude-device-files",
    "--exclude-filelist",
    "--exclude-if-present",
    "--exclude-older-than",
    "--exclude-other-filesystems",
    "--exclude-regexp",
    "--include",
    "--include-filelist",
    "--include-regexp",
    "--files-from",
    "--filter-globbing",
    "--filter-ignorecase",
    "--filter-literal",
    "--filter-regexp",
    "--filter-strictcase",
}

changed_options = {
    "--file-to-restore",
    "--do-not-restore-ownership",
}

removed_options = {
    "--gio",
    "--old-filenames",
    "--short-filenames",
    "--exclude-globbing-filelist",
    "--include-globbing-filelist",
    "--exclude-filelist-stdin",
    "--include-filelist-stdin",
    "--s3-multipart-max-timeout",
    "--s3-european-buckets",
    "--s3-use-multiprocessing",
    "--s3-use-new-style",
}

removed_backup_options = {
    "--time-separator",
}

# make list of all options available
all_options = {var2opt(var) for var in OptionKwargs.keys()}

trans = {
    # TRANSL: Used in usage help to represent a Unix-style path name. Example:
    # rsync://user[:password]@other_host[:port]//absolute_path
    "absolute_path": _("absolute_path"),
    # TRANSL: Used in usage help. Example:
    # tahoe://alias/some_dir
    "alias": _("alias"),
    # TRANSL: Used in help to represent a "bucket name" for Amazon Web
    # Services' Simple Storage Service (S3). Example:
    # s3://other.host/bucket_name[/prefix]
    "bucket_name": _("bucket_name"),
    # TRANSL: abbreviation for "character" (noun)
    "char": _("char"),
    # TRANSL: noun
    "command": _("command"),
    # TRANSL: Used in usage help to represent the name of a container in
    # Amazon Web Services' Cloudfront. Example:
    # cf+http://container_name
    "container_name": _("container_name"),
    # TRANSL: noun
    "count": _("count"),
    # TRANSL: Used in usage help to represent the name of a file directory
    "directory": _("directory"),
    # TRANSL: Used in usage help to represent the name of a file. Example:
    # --log-file <filename>
    "filename": _("filename"),
    # TRANSL: Used in usage help to represent an ID for a GnuPG key. Example:
    # --encrypt-key <gpg_key_id>
    "gpg_key_id": _("gpg-key-id"),
    # TRANSL: Used in usage help, e.g. to represent the name of a code
    # module. Example:
    # rsync://user[:password]@other.host[:port]::/module/some_dir
    "module": _("module"),
    # TRANSL: Used in usage help to represent a desired number of
    # something. Example:
    # --num-retries <number>
    "number": _("number"),
    # TRANSL: Used in usage help. (Should be consistent with the "Options:"
    # header.) Example:
    # duplicity [full|incremental] [options] source_path target_url
    "options": _("options"),
    # TRANSL: Used in usage help to represent an internet hostname. Example:
    # ftp://user[:password]@other.host[:port]/some_dir
    "other_host": _("other.host"),
    # TRANSL: Used in usage help. Example:
    # ftp://user[:password]@other.host[:port]/some_dir
    "password": _("password"),
    # TRANSL: Used in usage help to represent a Unix-style path name. Example:
    # --archive-dir <path>
    "path": _("path"),
    # TRANSL: Used in usage help to represent a TCP port number. Example:
    # ftp://user[:password]@other.host[:port]/some_dir
    "port": _("port"),
    # TRANSL: Used in usage help. This represents a string to be used as a
    # prefix to names for backup files created by Duplicity. Example:
    # s3://other.host/bucket_name[/prefix]
    "prefix": _("prefix"),
    # TRANSL: Used in usage help to represent a Unix-style path name. Example:
    # rsync://user[:password]@other.host[:port]/relative_path
    "relative_path": _("relative_path"),
    # TRANSL: Used in usage help. Example:
    # --timeout <seconds>
    "seconds": _("seconds"),
    # TRANSL: Used in usage help to represent a "glob" style pattern for
    # matching one or more files, as described in the documentation.
    # Example:
    # --exclude <shell_pattern>
    "shell_pattern": _("shell_pattern"),
    # TRANSL: Used in usage help to represent the name of a single file
    # directory or a Unix-style path to a directory. Example:
    # file:///some_dir
    "some_dir": _("some_dir"),
    # TRANSL: Used in usage help to represent the name of a single file
    # directory or a Unix-style path to a directory where files will be
    # coming FROM. Example:
    # duplicity [full|incremental] [options] source_path target_url
    "source_path": _("source_path"),
    # TRANSL: Used in usage help to represent a URL files will be coming
    # FROM. Example:
    # duplicity [restore] [options] source_url target_path
    "source_url": _("source_url"),
    # TRANSL: Used in usage help to represent the name of a single file
    # directory or a Unix-style path to a directory. where files will be
    # going TO. Example:
    # duplicity [restore] [options] source_url target_path
    "target_path": _("target_path"),
    # TRANSL: Used in usage help to represent a URL files will be going TO.
    # Example:
    # duplicity [full|incremental] [options] source_path target_url
    "target_url": _("target_url"),
    # TRANSL: Used in usage help to represent a time spec for a previous
    # point in time, as described in the documentation. Example:
    # duplicity remove-older-than time [options] target_url
    "time": _("time"),
    # TRANSL: Used in usage help to represent a user name (i.e. login).
    # Example:
    # ftp://user[:password]@other.host[:port]/some_dir
    "user": _("user"),
    # TRANSL: account id for b2. Example: b2://account_id@bucket/
    "account_id": _("account_id"),
    # TRANSL: application_key for b2.
    # Example: b2://account_id:application_key@bucket/
    "application_key": _("application_key"),
    # TRANSL: remote name for rclone.
    # Example: rclone://remote:/some_dir
    "remote": _("remote"),
}

help_url_formats = _("Backends and their URL formats:") + f"""
  azure://{trans['container_name']}
  b2://{trans['account_id']}[:{trans['application_key']}]@{trans['bucket_name']}/[{trans['some_dir']}/]
  boto3+s3://{trans['bucket_name']}[/{trans['prefix']}]
  cf+http://{trans['container_name']}
  dpbx:///{trans['some_dir']}
  file:///{trans['some_dir']}
  ftp://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]/{trans['some_dir']}
  ftps://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]/{trans['some_dir']}
  gdocs://{trans['user']}[:{trans['password']}]@{trans['other_host']}/{trans['some_dir']}
  for gdrive:// a <service-account-url> like the following is required
        <serviceaccount-name>@<serviceaccount-name>.iam.gserviceaccount.com
  gdrive://<service-account-url>/target-folder/?driveID=<SHARED DRIVE ID> (for GOOGLE Shared Drive)
  gdrive://<service-account-url>/target-folder/?myDriveFolderID=<google-myDrive-folder-id> (for GOOGLE MyDrive)
  hsi://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]/{trans['some_dir']}
  imap://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]/{trans['some_dir']}
  mega://{trans['user']}[:{trans['password']}]@{trans['other_host']}/{trans['some_dir']}
  megav2://{trans['user']}[:{trans['password']}]@{trans['other_host']}/{trans['some_dir']}
  mf://{trans['user']}[:{trans['password']}]@{trans['other_host']}/{trans['some_dir']}
  onedrive://{trans['some_dir']}
  pca://{trans['container_name']}
  pydrive://{trans['user']}@{trans['other_host']}/{trans['some_dir']}
  rclone://{trans['remote']}:/{trans['some_dir']}
  rsync://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]/{trans['relative_path']}
  rsync://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]//{trans['absolute_path']}
  rsync://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]::/{trans['module']}/{trans['some_dir']}
  s3+http://{trans['bucket_name']}[/{trans['prefix']}]
  s3://{trans['other_host']}[:{trans['port']}]/{trans['bucket_name']}[/{trans['prefix']}]
  scp://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]/{trans['some_dir']}
  ssh://{trans['user']}[:{trans['password']}]@{trans['other_host']}[:{trans['port']}]/{trans['some_dir']}
  swift://{trans['container_name']}
  tahoe://{trans['alias']}/{trans['directory']}
  webdav://{trans['user']}[:{trans['password']}]@{trans['other_host']}/{trans['some_dir']}
  webdavs://{trans['user']}[:{trans['password']}]@{trans['other_host']}/{trans['some_dir']}
"""



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2008 Francois Deppierraz
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Tahoe-LAFS backend.
"""

import os

import duplicity.backend
from duplicity import log


class TAHOEBackend(duplicity.backend.Backend):
    """
    Backend for the Tahoe file system
    """

    def __init__(self, parsed_url):
        """
        Initialize the Tahoe backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        url = parsed_url.path.strip("/").split("/")

        self.alias = url[0]

        if len(url) > 1:
            self.directory = "/".join(url[1:])
        else:
            self.directory = ""

        log.Debug(f"tahoe: {url} -> {self.alias}:{self.directory}")

    def get_remote_path(self, filename=None):
        if filename is None:
            if self.directory != "":
                return f"{self.alias}:{self.directory}"
            else:
                return f"{self.alias}:"

        if isinstance(filename, bytes):
            filename = os.fsdecode(filename)
        if self.directory != "":
            return f"{self.alias}:{self.directory}/{filename}"
        else:
            return f"{self.alias}:{filename}"

    def run(self, *args):
        cmd = " ".join(args)
        _, output, _ = self.subprocess_popen(cmd)
        return output

    def _put(self, source_path, remote_filename):
        self.run("tahoe", "cp", source_path.uc_name, self.get_remote_path(remote_filename))

    def _get(self, remote_filename, local_path):
        self.run("tahoe", "cp", self.get_remote_path(remote_filename), local_path.uc_name)

    def _list(self):
        output = self.run("tahoe", "ls", self.get_remote_path())
        return [os.fsencode(x) for x in output.split("\n") if x]

    def _delete(self, filename):
        self.run("tahoe", "rm", self.get_remote_path(filename))


duplicity.backend.register_backend("tahoe", TAHOEBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2015 Yigal Asnis
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# It is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
PyDrive backend (deprecated).
"""

import os

import duplicity.backend
from duplicity import errors
from duplicity import log
from duplicity.errors import BackendException


class PyDriveBackend(duplicity.backend.Backend):
    """
    Connect to remote store using PyDrive API.
    """

    def __init__(self, parsed_url):
        """
        Initialize the PyDrive backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        log.Warn(
            "\n\nThe PyDrive backend is now deprecated and will be removed in a future version.\n"
            "It has been replaced by the GDrive backend.  See the manpage for details.\n\n\n"
        )

        try:
            import httplib2
            from apiclient.discovery import build
        except ImportError as e:
            raise BackendException(
                "PyDrive backend requires PyDrive2 and Google API client installation.\n"
                "Please read the manpage for setup details.\n"
                f"Exception: {str(e)}"
            )

        # Shared Drive ID specified as a query parameter in the backend URL.
        # Example: pydrive://developer.gserviceaccount.com/target-folder/?driveID=<SHARED DRIVE ID>
        self.api_params = {}
        self.shared_drive_id = None
        if "driveID" in parsed_url.query_args:
            self.shared_drive_id = parsed_url.query_args["driveID"][0]
            self.api_params = {
                "corpora": "teamDrive",
                "teamDriveId": self.shared_drive_id,
                "includeTeamDriveItems": True,
                "supportsTeamDrives": True,
            }

        try:
            from pydrive2.auth import GoogleAuth
            from pydrive2.drive import GoogleDrive
            from pydrive2.files import ApiRequestError
            from pydrive2.files import FileNotUploadedError
        except ImportError as e:
            raise BackendException(
                f"PyDrive backend requires PyDrive2 installation.\n"
                f"Please read the manpage for setup details.\n"
                f"Exception: {str(e)}"
            )

        # let user get by with old client while he can
        try:
            from oauth2client.client import SignedJwtAssertionCredentials

            self.oldClient = True
        except Exception as e:
            from oauth2client.service_account import ServiceAccountCredentials
            from oauth2client import crypt

            self.oldClient = False

        if "GOOGLE_DRIVE_ACCOUNT_KEY" in os.environ:
            account_key = os.environ["GOOGLE_DRIVE_ACCOUNT_KEY"]
            if self.oldClient:
                credentials = SignedJwtAssertionCredentials(
                    f"{parsed_url.username}@{parsed_url.hostname}",
                    account_key,
                    scopes="https://www.googleapis.com/auth/drive",
                )
            else:
                signer = crypt.Signer.from_string(account_key)  # pylint: disable=used-before-assignment
                credentials = ServiceAccountCredentials(  # pylint: disable=used-before-assignment
                    f"{parsed_url.username}@{parsed_url.hostname}",
                    signer,
                    scopes="https://www.googleapis.com/auth/drive",
                )
            credentials.authorize(httplib2.Http())
            gauth = GoogleAuth(http_timeout=60)
            gauth.credentials = credentials
        elif "GOOGLE_DRIVE_SETTINGS" in os.environ:
            gauth = GoogleAuth(settings_file=os.environ["GOOGLE_DRIVE_SETTINGS"], http_timeout=60)
            gauth.CommandLineAuth()
        elif "GOOGLE_SECRETS_FILE" in os.environ and "GOOGLE_CREDENTIALS_FILE" in os.environ:
            gauth = GoogleAuth(http_timeout=60)
            gauth.LoadClientConfigFile(os.environ["GOOGLE_SECRETS_FILE"])
            gauth.LoadCredentialsFile(os.environ["GOOGLE_CREDENTIALS_FILE"])
            if gauth.credentials is None:
                gauth.CommandLineAuth()
            elif gauth.access_token_expired:
                gauth.Refresh()
            else:
                gauth.Authorize()
            gauth.SaveCredentialsFile(os.environ["GOOGLE_CREDENTIALS_FILE"])
        else:
            raise BackendException(
                "GOOGLE_DRIVE_ACCOUNT_KEY or GOOGLE_DRIVE_SETTINGS environment "
                "variable not set. Please read the manpage to fix."
            )
        self.drive = GoogleDrive(gauth)

        if self.shared_drive_id:
            parent_folder_id = self.shared_drive_id
        else:
            # Dirty way to find root folder id
            file_list = self.drive.ListFile({"q": "'Root' in parents and trashed=false"}).GetList()
            if file_list:
                parent_folder_id = file_list[0]["parents"][0]["id"]
            else:
                file_in_root = self.drive.CreateFile({"title": "i_am_in_root"})
                file_in_root.Upload()
                parent_folder_id = file_in_root["parents"][0]["id"]
                file_in_root.Delete()

        # Fetch destination folder entry and create hierarchy if required.
        folder_names = parsed_url.path.split("/")
        for folder_name in folder_names:
            if not folder_name:
                continue
            list_file_args = {"q": f"'{parent_folder_id}' in parents and trashed=false"}
            list_file_args.update(self.api_params)
            file_list = self.drive.ListFile(list_file_args).GetList()
            folder = next(
                (
                    item
                    for item in file_list
                    if item["title"] == folder_name and item["mimeType"] == "application/vnd.google-apps.folder"
                ),
                None,
            )
            if folder is None:
                create_file_args = {
                    "title": folder_name,
                    "mimeType": "application/vnd.google-apps.folder",
                    "parents": [{"id": parent_folder_id}],
                }
                create_file_args["parents"][0].update(self.api_params)
                create_file_args.update(self.api_params)
                folder = self.drive.CreateFile(create_file_args)
                if self.shared_drive_id:
                    folder.Upload(param={"supportsTeamDrives": True})
                else:
                    folder.Upload()
            parent_folder_id = folder["id"]
        self.folder = parent_folder_id
        self.id_cache = {}

    def file_by_name(self, filename):
        from pydrive2.files import ApiRequestError  # pylint: disable=import-error

        filename = os.fsdecode(filename)  # PyDrive deals with unicode filenames

        if filename in self.id_cache:
            # It might since have been locally moved, renamed or deleted, so we
            # need to validate the entry.
            file_id = self.id_cache[filename]
            drive_file = self.drive.CreateFile({"id": file_id})
            try:
                if drive_file["title"] == filename and not drive_file["labels"]["trashed"]:
                    for parent in drive_file["parents"]:
                        if parent["id"] == self.folder:
                            log.Info(f"PyDrive backend: found file '{filename}' with id {file_id} in ID cache")
                            return drive_file
            except ApiRequestError as error:
                # A 404 occurs if the ID is no longer valid
                if error.args[0].resp.status != 404:
                    raise
            # If we get here, the cache entry is invalid
            log.Info(f"PyDrive backend: invalidating '{filename}' (previously ID {file_id}) from ID cache")
            del self.id_cache[filename]

        # Not found in the cache, so use directory listing. This is less
        # reliable because there is no strong consistency.
        q = f"title='{filename}' and '{self.folder}' in parents and trashed=false"
        fields = "items(title,id,fileSize,downloadUrl,exportLinks),nextPageToken"
        list_file_args = {"q": q, "fields": fields}
        list_file_args.update(self.api_params)
        flist = self.drive.ListFile(list_file_args).GetList()
        if len(flist) > 1:
            raise errors.FatalError(_("PyDrive backend: multiple files called '%s'.") % (filename,))
        elif flist:
            file_id = flist[0]["id"]
            self.id_cache[filename] = flist[0]["id"]
            log.Info(f"PyDrive backend: found file '{filename}' with id {file_id} on server, adding to cache")
            return flist[0]
        log.Info(f"PyDrive backend: file '{filename}' not found in cache or on server")
        return None

    def id_by_name(self, filename):
        drive_file = self.file_by_name(filename)
        if drive_file is None:
            return ""
        else:
            return drive_file["id"]

    def _put(self, source_path, remote_filename):
        remote_filename = os.fsdecode(remote_filename)
        drive_file = self.file_by_name(remote_filename)
        if drive_file is None:
            # No existing file, make a new one
            create_file_args = {
                "title": remote_filename,
                "parents": [{"kind": "drive#fileLink", "id": self.folder}],
            }
            create_file_args["parents"][0].update(self.api_params)
            drive_file = self.drive.CreateFile(create_file_args)
            log.Info(f"PyDrive backend: creating new file '{remote_filename}'")
        else:
            log.Info(f"PyDrive backend: replacing existing file '{remote_filename}' with id '{drive_file['id']}'")
        drive_file.SetContentFile(os.fsdecode(source_path.name))
        if self.shared_drive_id:
            drive_file.Upload(param={"supportsTeamDrives": True})
        else:
            drive_file.Upload()
        self.id_cache[remote_filename] = drive_file["id"]

    def _get(self, remote_filename, local_path):
        drive_file = self.file_by_name(remote_filename)
        drive_file.GetContentFile(os.fsdecode(local_path.name))

    def _list(self):
        list_file_args = {
            "q": f"'{self.folder}' in parents and trashed=false",
            "fields": "items(title,id),nextPageToken",
        }
        list_file_args.update(self.api_params)
        drive_files = self.drive.ListFile(list_file_args).GetList()
        filenames = set(item["title"] for item in drive_files)
        # Check the cache as well. A file might have just been uploaded but
        # not yet appear in the listing.
        # Note: do not use iterkeys() here, because file_by_name will modify
        # the cache if it finds invalid entries.
        for filename in list(self.id_cache.keys()):
            if (filename not in filenames) and (self.file_by_name(filename) is not None):
                filenames.add(filename)
        return list(filenames)

    def _delete(self, filename):
        file_id = self.id_by_name(filename)
        if file_id == "":
            log.Warn(f"File '{os.fsdecode(filename)}' does not exist while trying to delete it")
        elif self.shared_drive_id:
            self.drive.auth.service.files().delete(fileId=file_id, param={"supportsTeamDrives": True}).execute()
        else:
            self.drive.auth.service.files().delete(fileId=file_id).execute()

    def _query(self, filename):
        drive_file = self.file_by_name(filename)
        if drive_file is None:
            size = -1
        else:
            size = int(drive_file["fileSize"])
        return {"size": size}

    def _error_code(self, operation, error):  # pylint: disable=unused-argument
        from pydrive2.files import ApiRequestError
        from pydrive2.files import FileNotUploadedError  # pylint: disable=import-error

        if isinstance(error, FileNotUploadedError):
            return log.ErrorCode.backend_not_found
        elif isinstance(error, ApiRequestError):
            return log.ErrorCode.backend_permission_denied
        return log.ErrorCode.backend_error


duplicity.backend.register_backend("pydrive", PyDriveBackend)
"""
pydrive is an alternate way to access gdocs.
"""
duplicity.backend.register_backend("pydrive+gdocs", PyDriveBackend)
"""
register pydrive as the default way to access gdocs.
"""
duplicity.backend.register_backend("gdocs", PyDriveBackend)

duplicity.backend.uses_netloc.extend(["pydrive", "pydrive+gdocs", "gdocs"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2022 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
ctypes-based interface to librsync.
"""

import ctypes
import ctypes.util
import os

# Load librsync
_librsync_path = ctypes.util.find_library("rsync")
if not _librsync_path:
    # Try common locations if find_library fails
    for path in ["librsync.so", "librsync.dylib", "librsync.so.2", "librsync.so.1"]:
        try:
            ctypes.CDLL(path)
            _librsync_path = path
            break
        except OSError:
            continue

if not _librsync_path:
    raise ImportError("librsync library not found")

_librsync = ctypes.CDLL(_librsync_path)

# Constants
RS_JOB_BLOCKSIZE = 65536
RS_DEFAULT_BLOCK_LEN = 2048

# Result codes
RS_DONE = 0
RS_BLOCKED = 1

# Magic numbers
RS_MD4_SIG_MAGIC = 0x72730136
RS_BLAKE2_SIG_MAGIC = 0x72730137


class librsyncError(Exception):
    """
    Signifies an error in internal librsync processing.
    """

    pass


# Structures
class rs_buffers_t(ctypes.Structure):
    """
    ctypes structure for rs_buffers_t.
    """

    _fields_ = [
        ("next_in", ctypes.c_char_p),
        ("avail_in", ctypes.c_size_t),
        ("eof_in", ctypes.c_int),
        ("next_out", ctypes.c_char_p),
        ("avail_out", ctypes.c_size_t),
    ]


# Function prototypes
_librsync.rs_sig_begin.restype = ctypes.c_void_p
_librsync.rs_sig_begin.argtypes = [ctypes.c_size_t, ctypes.c_size_t, ctypes.c_int]

_librsync.rs_loadsig_begin.restype = ctypes.c_void_p
_librsync.rs_loadsig_begin.argtypes = [ctypes.POINTER(ctypes.c_void_p)]

_librsync.rs_delta_begin.restype = ctypes.c_void_p
_librsync.rs_delta_begin.argtypes = [ctypes.c_void_p]

_librsync.rs_patch_begin.restype = ctypes.c_void_p
# rs_copy_cb(void *opaque, rs_long_t pos, size_t *len, void **buf)
RS_COPY_CB = ctypes.CFUNCTYPE(
    ctypes.c_int, ctypes.c_void_p, ctypes.c_longlong, ctypes.POINTER(ctypes.c_size_t), ctypes.POINTER(ctypes.c_void_p)
)
_librsync.rs_patch_begin.argtypes = [RS_COPY_CB, ctypes.c_void_p]

_librsync.rs_job_iter.restype = ctypes.c_int
_librsync.rs_job_iter.argtypes = [ctypes.c_void_p, ctypes.POINTER(rs_buffers_t)]

_librsync.rs_job_free.restype = ctypes.c_int
_librsync.rs_job_free.argtypes = [ctypes.c_void_p]

_librsync.rs_build_hash_table.restype = ctypes.c_int
_librsync.rs_build_hash_table.argtypes = [ctypes.c_void_p]

_librsync.rs_free_sumset.restype = None
_librsync.rs_free_sumset.argtypes = [ctypes.c_void_p]

_librsync.rs_strerror.restype = ctypes.c_char_p
_librsync.rs_strerror.argtypes = [ctypes.c_int]


def _set_error(result, location):
    err_msg = _librsync.rs_strerror(result).decode("utf-8")
    raise librsyncError(f"librsync error {result} ({err_msg}) while in {location}")


class SigMaker:
    """
    Class to generate signatures using librsync.
    """

    def __init__(self, blocklen):
        """
        Initialize the SigMaker.

        :param int blocklen: The block length for the signature.
        """
        # Using 8 for strong_len and RS_MD4_SIG_MAGIC as in _librsyncmodule.c
        self.job = _librsync.rs_sig_begin(blocklen, 8, RS_MD4_SIG_MAGIC)
        if not self.job:
            raise librsyncError("Failed to create sig_job")

    def cycle(self, inbuf):
        """
        Process a chunk of input data for signature generation.

        :param bytes inbuf: The input data to process.
        :returns: A tuple of (done, bytes_read, cycle_out).
        :rtype: tuple
        """
        buf = rs_buffers_t()
        buf.next_in = ctypes.c_char_p(inbuf)
        buf.avail_in = len(inbuf)
        buf.eof_in = len(inbuf) == 0

        outbuf = ctypes.create_string_buffer(RS_JOB_BLOCKSIZE)
        buf.next_out = ctypes.cast(outbuf, ctypes.c_char_p)
        buf.avail_out = RS_JOB_BLOCKSIZE

        result = _librsync.rs_job_iter(self.job, ctypes.byref(buf))

        if result not in [RS_DONE, RS_BLOCKED]:
            _set_error(result, "sig cycle")

        return (result == RS_DONE), len(inbuf) - buf.avail_in, outbuf.raw[: RS_JOB_BLOCKSIZE - buf.avail_out]

    def __del__(self):
        if hasattr(self, "job") and self.job:
            _librsync.rs_job_free(self.job)


class DeltaMaker:
    """
    Class to generate deltas using librsync.
    """

    def __init__(self, sig_string):
        """
        Initialize the DeltaMaker.

        :param bytes sig_string: The signature to use for delta generation.
        """
        self.sig_ptr = ctypes.c_void_p()
        sig_loader = _librsync.rs_loadsig_begin(ctypes.byref(self.sig_ptr))

        buf = rs_buffers_t()
        buf.next_in = ctypes.c_char_p(sig_string)
        buf.avail_in = len(sig_string)
        buf.eof_in = 1

        outbuf = ctypes.create_string_buffer(RS_JOB_BLOCKSIZE)
        buf.next_out = ctypes.cast(outbuf, ctypes.c_char_p)
        buf.avail_out = RS_JOB_BLOCKSIZE

        result = _librsync.rs_job_iter(sig_loader, ctypes.byref(buf))
        _librsync.rs_job_free(sig_loader)

        if result != RS_DONE:
            _set_error(result, "delta rs_signature_t builder")

        result = _librsync.rs_build_hash_table(self.sig_ptr)
        if result != RS_DONE:
            _set_error(result, "delta rs_build_hash_table")

        self.job = _librsync.rs_delta_begin(self.sig_ptr)
        if not self.job:
            raise librsyncError("Failed to create delta_job")

    def cycle(self, inbuf):
        """
        Process a chunk of input data for delta generation.

        :param bytes inbuf: The input data to process.
        :returns: A tuple of (done, bytes_read, cycle_out).
        :rtype: tuple
        """
        buf = rs_buffers_t()
        buf.next_in = ctypes.c_char_p(inbuf)
        buf.avail_in = len(inbuf)
        buf.eof_in = len(inbuf) == 0

        outbuf = ctypes.create_string_buffer(RS_JOB_BLOCKSIZE)
        buf.next_out = ctypes.cast(outbuf, ctypes.c_char_p)
        buf.avail_out = RS_JOB_BLOCKSIZE

        result = _librsync.rs_job_iter(self.job, ctypes.byref(buf))

        if result not in [RS_DONE, RS_BLOCKED]:
            _set_error(result, "delta cycle")

        return (result == RS_DONE), len(inbuf) - buf.avail_in, outbuf.raw[: RS_JOB_BLOCKSIZE - buf.avail_out]

    def __del__(self):
        if hasattr(self, "job") and self.job:
            _librsync.rs_job_free(self.job)
        if hasattr(self, "sig_ptr") and self.sig_ptr:
            _librsync.rs_free_sumset(self.sig_ptr)


class PatchMaker:
    """
    Class to apply patches using librsync.
    """

    def __init__(self, basis_file):
        """
        Initialize the PatchMaker.

        :param basis_file: The original file to patch.
        """
        self.basis_file = basis_file
        # We need to keep a reference to the callback to prevent it from being GC'd
        self._callback = RS_COPY_CB(self._copy_cb)
        self.job = _librsync.rs_patch_begin(self._callback, None)
        self._internal_buf = None

    def _copy_cb(self, opaque, pos, length_ptr, buf_ptr):
        try:
            self.basis_file.seek(pos)
            data = self.basis_file.read(length_ptr[0])
            # We must provide a buffer that stays alive until librsync is done with it.
            # rs_job_iter for patch will call this and then consume the buffer.
            self._internal_buf = ctypes.create_string_buffer(data)
            buf_ptr[0] = ctypes.cast(self._internal_buf, ctypes.c_void_p)
            length_ptr[0] = len(data)
            return RS_DONE
        except Exception:
            return -1  # RS_IO_ERROR or similar

    def cycle(self, inbuf):
        """
        Process a chunk of input data for patching.

        :param bytes inbuf: The input data to process.
        :returns: A tuple of (done, bytes_read, cycle_out).
        :rtype: tuple
        """
        buf = rs_buffers_t()
        buf.next_in = ctypes.c_char_p(inbuf)
        buf.avail_in = len(inbuf)
        buf.eof_in = len(inbuf) == 0

        outbuf = ctypes.create_string_buffer(RS_JOB_BLOCKSIZE)
        buf.next_out = ctypes.cast(outbuf, ctypes.c_char_p)
        buf.avail_out = RS_JOB_BLOCKSIZE

        result = _librsync.rs_job_iter(self.job, ctypes.byref(buf))

        if result not in [RS_DONE, RS_BLOCKED]:
            _set_error(result, "patch cycle")

        return (result == RS_DONE), len(inbuf) - buf.avail_in, outbuf.raw[: RS_JOB_BLOCKSIZE - buf.avail_out]

    def __del__(self):
        if hasattr(self, "job") and self.job:
            _librsync.rs_job_free(self.job)


def new_sigmaker(blocklen):
    """
    Create a new SigMaker instance.

    :param int blocklen: The block length.
    :returns: A new SigMaker instance.
    :rtype: SigMaker
    """
    return SigMaker(blocklen)


def new_deltamaker(sig_string):
    """
    Create a new DeltaMaker instance.

    :param bytes sig_string: The signature string.
    :returns: A new DeltaMaker instance.
    :rtype: DeltaMaker
    """
    return DeltaMaker(sig_string)


def new_patchmaker(basis_file):
    """
    Create a new PatchMaker instance.

    :param basis_file: The original file to patch.
    :returns: A new PatchMaker instance.
    :rtype: PatchMaker
    """
    return PatchMaker(basis_file)


def get_version():
    """
    Return the version string of the loaded librsync library.

    :returns: The version string.
    :rtype: str
    """
    return ctypes.c_char_p(ctypes.addressof(ctypes.c_char.in_dll(_librsync, "rs_librsync_version"))).value.decode(
        "utf-8"
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
SSH backend using pexpect.
"""

# The following can be redefined to use different shell commands from
# ssh or scp or to add more arguments.  However, the replacements must
# have the same syntax.  Also these strings will be executed by the
# shell, so shouldn't have strange characters in them.


import os
import re

import duplicity.backend
from duplicity import config
from duplicity import errors
from duplicity import log
from duplicity.errors import BackendException


class SSHPExpectBackend(duplicity.backend.Backend):
    """This backend copies files using scp.  List not supported.  Filenames
    should not need any quoting or this will break."""

    def __init__(self, parsed_url):
        """
        Initialize the SSH pexpect backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        try:
            global pexpect
            import pexpect
        except ImportError:
            raise

        if pexpect.__version__ < "4.5.0":
            raise errors.FatalError(f"""
                The version of pexpect, '{pexpect.__version__}', is too old.  We need version 4.5.0 or above to run.
                See https://gitlab.com/duplicity/duplicity/-/issues/125 for the gory details.

                Use "python3 -m pip install pexpect" to install the latest version.
                """)

        self.retry_delay = 10

        self.scp_command = "scp"
        if config.scp_command:
            self.scp_command = config.scp_command

        self.sftp_command = "sftp"
        if config.sftp_command:
            self.sftp_command = config.sftp_command

        self.scheme = duplicity.backend.strip_prefix(parsed_url.scheme, "pexpect")
        self.use_scp = self.scheme == "scp"

        # host string of form [user@]hostname
        if parsed_url.username:
            self.host_string = f"{parsed_url.username}@{parsed_url.hostname}"
        else:
            self.host_string = parsed_url.hostname
        # make sure remote_dir is always valid
        if parsed_url.path:
            # remove leading '/'
            self.remote_dir = re.sub(r"^/", r"", parsed_url.path, 1)
        else:
            self.remote_dir = "."
        self.remote_prefix = f"{self.remote_dir}/"
        # maybe use different ssh port
        if parsed_url.port:
            config.ssh_options = f"{config.ssh_options} -oPort={parsed_url.port}"
        # set some defaults if user has not specified already.
        if "ServerAliveInterval" not in config.ssh_options:
            config.ssh_options += f" -oServerAliveInterval={int(int(config.timeout / 2))}"
        if "ServerAliveCountMax" not in config.ssh_options:
            config.ssh_options += " -oServerAliveCountMax=2"

        # set up password
        self.use_getpass = config.ssh_askpass
        self.password = self.get_password()

    def run_scp_command(self, commandline):
        """
        Run an scp command, responding to password prompts
        """
        log.Info(f"Running '{commandline}'")
        child = pexpect.spawn(commandline, timeout=None, use_poll=True)
        if config.ssh_askpass:
            state = "authorizing"
        else:
            state = "copying"
        while True:
            if state == "authorizing":
                match = child.expect(
                    [
                        pexpect.EOF,
                        "(?i)timeout, server not responding",
                        "(?i)pass(word|phrase .*):",
                        "(?i)permission denied",
                        "authenticity",
                    ]
                )
                log.Debug(f"State = {state}, Before = '{child.before.strip()}'")
                if match == 0:
                    log.Warn("Failed to authenticate")
                    break
                elif match == 1:
                    log.Warn("Timeout waiting to authenticate")
                    break
                elif match == 2:
                    child.sendline(self.password)
                    state = "copying"
                elif match == 3:
                    log.Warn("Invalid SSH password")
                    break
                elif match == 4:
                    log.Warn("Remote host authentication failed (missing known_hosts entry?)")
                    break
            elif state == "copying":
                match = child.expect(
                    [
                        pexpect.EOF,
                        "(?i)timeout, server not responding",
                        "stalled",
                        "authenticity",
                        "ETA",
                    ]
                )
                log.Debug(f"State = {state}, Before = '{child.before.strip()}'")
                if match == 0:
                    break
                elif match == 1:
                    log.Warn("Timeout waiting for response")
                    break
                elif match == 2:
                    state = "stalled"
                elif match == 3:
                    log.Warn("Remote host authentication failed (missing known_hosts entry?)")
                    break
            elif state == "stalled":
                match = child.expect([pexpect.EOF, "(?i)timeout, server not responding", "ETA"])
                log.Debug(f"State = {state}, Before = '{child.before.strip()}'")
                if match == 0:
                    break
                elif match == 1:
                    log.Warn("Stalled for too long, aborted copy")
                    break
                elif match == 2:
                    state = "copying"
        child.close(force=True)
        if child.exitstatus != 0:
            raise BackendException(f"Error running '{commandline}'")

    def run_sftp_command(self, commandline, commands):
        """
        Run an sftp command, responding to password prompts, passing commands from list
        """
        maxread = 2000  # expected read buffer size
        responses = [
            pexpect.EOF,
            "(?i)timeout, server not responding",
            "sftp>",
            "(?i)pass(word|phrase .*):",
            "(?i)permission denied",
            "authenticity",
            "(?i)no such file or directory",
            "Couldn't delete file: No such file or directory",
            "Couldn't delete file",
            "open(.*): Failure",
        ]
        max_response_len = max([len(p) for p in responses[1:]])
        log.Info(f"Running '{commandline}'")
        child = pexpect.spawn(
            commandline,
            timeout=None,
            maxread=maxread,
            encoding=config.fsencoding,
            use_poll=True,
        )
        cmdloc = 0
        passprompt = 0
        while True:
            msg = ""
            match = child.expect(responses, searchwindowsize=maxread + max_response_len)
            log.Debug(f"State = sftp, Before = '{child.before.strip()}'")
            if match == 0:
                break
            elif match == 1:
                msg = "Timeout waiting for response"
                break
            if match == 2:
                if cmdloc < len(commands):
                    command = commands[cmdloc]
                    log.Info(f"sftp command: '{command}'")
                    child.sendline(command)
                    cmdloc += 1
                else:
                    command = "quit"
                    child.sendline(command)
                    res = child.before
            elif match == 3:
                passprompt += 1
                child.sendline(self.password)
                if passprompt > 1:
                    raise BackendException("Invalid SSH password.")
            elif match == 4:
                if not child.before.strip().startswith("mkdir"):
                    msg = "Permission denied"
                    break
            elif match == 5:
                msg = "Host key authenticity could not be verified (missing known_hosts entry?)"
                break
            elif match == 6:
                if not child.before.strip().startswith("rm"):
                    msg = f"Remote file or directory does not exist in command='{commandline}'"
                    break
            elif match == 7:
                if not child.before.strip().startswith("Removing"):
                    msg = f"Could not delete file in command='{commandline}'"
                    break
            elif match == 8:
                msg = f"Could not delete file in command='{commandline}'"
                break
            elif match == 9:
                msg = f"Could not open file in command='{commandline}'"
                break
        child.close(force=True)
        if child.exitstatus == 0:
            return res
        else:
            raise BackendException(f"Error running '{commandline}': {msg}")

    def _put(self, source_path, remote_filename):
        remote_filename = os.fsdecode(remote_filename)
        if self.use_scp:
            self.put_scp(source_path, remote_filename)
        else:
            self.put_sftp(source_path, remote_filename)

    def put_sftp(self, source_path, remote_filename):
        commands = [
            f'put "{source_path.uc_name}" "{self.remote_prefix}.{remote_filename}.part"',
            f'rename "{self.remote_prefix}.{remote_filename}.part" "{self.remote_prefix}{remote_filename}"',
        ]
        commandline = f"{self.sftp_command} {config.ssh_options} {self.host_string}"
        self.run_sftp_command(commandline, commands)

    def put_scp(self, source_path, remote_filename):
        commandline = (
            f"{self.scp_command} {config.ssh_options} {source_path.uc_name} "
            f"{self.host_string}:{self.remote_prefix}{remote_filename}"
        )
        self.run_scp_command(commandline)

    def _get(self, remote_filename, local_path):
        remote_filename = os.fsdecode(remote_filename)
        if self.use_scp:
            self.get_scp(remote_filename, local_path)
        else:
            self.get_sftp(remote_filename, local_path)

    def get_sftp(self, remote_filename, local_path):
        commands = [f'get "{self.remote_prefix}{remote_filename}" "{local_path.uc_name}"']
        commandline = f"{self.sftp_command} {config.ssh_options} {self.host_string}"
        self.run_sftp_command(commandline, commands)

    def get_scp(self, remote_filename, local_path):
        commandline = (
            f"{self.scp_command} "
            f"{config.ssh_options} "
            f"{self.host_string}:{self.remote_prefix}{remote_filename} "
            f"{local_path.uc_name}"
        )
        self.run_scp_command(commandline)

    def _list(self):
        """
        List files in the remote SSH folder using SFTP via pexpect.

        :returns: A list of filenames.
        :rtype: list
        """
        # Note that this command can get confused when dealing with
        # files with newlines in them, as the embedded newlines cannot
        # be distinguished from the file boundaries.
        dirs = self.remote_dir.split(os.sep)
        if len(dirs) > 0:
            if dirs[0] == "":
                dirs[0] = "/"
        mkdir_commands = []
        for d in dirs:
            mkdir_commands += [f'mkdir "{d}"'] + [f'cd "{d}"']

        commands = mkdir_commands + ["ls -1"]
        commandline = f"{self.sftp_command} {config.ssh_options} {self.host_string}"

        l = self.run_sftp_command(commandline, commands).split("\n")[1:]

        return [x for x in map(str.strip, l) if x]

    def _delete(self, filename):
        commands = [f'cd "{self.remote_dir}"']
        commands.append(f'rm "{os.fsdecode(filename)}"')
        commandline = f"{self.sftp_command} {config.ssh_options} {self.host_string}"
        self.run_sftp_command(commandline, commands)

    def _delete_list(self, filename_list):
        commands = [f'cd "{self.remote_dir}"']
        for filename in filename_list:
            commands.append(f'rm "{os.fsdecode(filename)}"')
        commandline = f"{self.sftp_command} {config.ssh_options} {self.host_string}"
        self.run_sftp_command(commandline, commands)


duplicity.backend.register_backend("pexpect+sftp", SSHPExpectBackend)
duplicity.backend.register_backend("pexpect+scp", SSHPExpectBackend)
duplicity.backend.uses_netloc.extend(["pexpect+sftp", "pexpect+scp"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Store global configuration information
"""

import os
import pickle
import socket
import sys

from duplicity import __version__, log
from duplicity import gpg

# The current version of duplicity
version = __version__

# The following args are set by commandline processing
# they correspond to the args in cli_main.duplicity_commands
count = None
remove_time = None
source_path = None
source_url = None
target_path = None
target_url = None

# action to take
action = None

# True if inc implied, not explicit
implied_inc = False

# Prefix for all files (appended before type-specific prefixes)
file_prefix = b""

# Prefix for manifest files only
file_prefix_manifest = b""

# Prefix for archive files only
file_prefix_archive = b""

# Prefix for sig files only
file_prefix_signature = b""

# Prefix for jsonstat files only
file_prefix_jsonstat = b""

# The name of the current host
hostname = socket.gethostname()

# For historical reasons also save the FQDN for comparing manifests, but
# we tend to prefer the hostname going forward.
fqdn = socket.getfqdn()

# The main local path.  For backing up the is the path to be backed
# up.  For restoring, this is the destination of the restored files.
local_path = None

# The symbolic name of the backup being operated upon.
backup_name = None

# Set to the Path of the archive directory (the directory which
# contains the signatures and manifests of the relevent backup
# collection), and for checkpoint state between volumes.
# NOTE: this gets expanded in duplicity.commandline
os.environ["XDG_CACHE_HOME"] = os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))
archive_dir = os.path.expandvars("$XDG_CACHE_HOME/duplicity")
archive_dir_path = None

# config dir for future use
os.environ["XDG_CONFIG_HOME"] = os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config"))
config_dir = os.path.expandvars("$XDG_CONFIG_HOME/duplicity")
config_dir_path = None

# Restores will try to bring back the state as of the following time.
# If it is None, default to current time.
restore_time = None

# If set, restore only the subdirectory or file specified, not the
# whole root.
restore_path = None

# The backend representing the remote side
backend = None

# Select object which iterates paths in the local source dir.
select = None
select_opts = []
select_files = []

# true if missing difftars should be run
# mmising difftar files fount in second call
first_colstats = False

# gpg binary to use
gpg_binary = None

# Options to pass to gpg
gpg_options = ""

# Set to GPGProfile that will be used to compress/uncompress encrypted
# files.  Replaces encryption_keys, sign_key, and passphrase settings.
gpg_profile = None

# Use gpgsm tool, which is very similar gpg but using PKCS#7 file format.
use_gpgsm = False

# Maximum file blocksize
max_blocksize = 0

# If true, filelists and directory statistics will be split on
# nulls instead of newlines.
null_separator = None

# number of retries on network operations
num_retries = 5

# True if Pydev debugger should be activated
pydevd = False

# Character used like the ":" in time strings like
# 2002-08-06T04:22:00-07:00.  The colon isn't good for filenames on
# windows machines.
time_separator = ":"

# Global lockfile used to manage concurrency
lockfile = None
lockpath = ""

# If this is true, only warn and don't raise fatal error when backup
# source directory doesn't match previous backup source directory.
allow_source_mismatch = False

# If set, print the statistics after every backup session
print_statistics = True

# If set, write extra statistic in json format to file
jsonstat = False

# If set, forces a full backup if the last full backup is older than
# the time specified
full_if_older_than = None

# If set, forces a full backup if the latest chain has N or more
# incremental backups
full_if_n_inc = None

# If set the incremental backup will be skipped if not DeltaEntries are detected
skip_if_no_change = False
# Track if incremental backup was skipped, no cli option
skipped_inc = False

# Used to confirm certain destructive operations like deleting old files.
force = None

# If set, signifies the number of backups chains to keep when performing
# a remove-all-but-n-full.
keep_chains = None

# Don't actually do anything, but still report what would be done
dry_run = False

# Compress files on remote system?
compression = True

# Encrypt files on remote system?
encryption = True

# volume size. default 200M
volsize = 200 * 1024 * 1024

# file copy blocksize
copy_blocksize = 128 * 1024

# Swift has a limit on the size of a single uploaded object; by default this is 5GB.
# https://docs.openstack.org/swift/latest/overview_large_objects.html
# With a volume large than this size, we will switch to multipart upload.
mp_segment_size = 5 * 2**30

# Working directory for the tempfile module. Defaults to /tmp on most systems.
temproot = None

# network timeout value
timeout = 30

# FTP data connection type
ftp_connection = "passive"

# Header options for Webdav
webdav_headers = ""

# Asynchronous put/get concurrency limit
# (default of 0 disables asynchronicity).
async_concurrency = 0

# File owner uid keeps number from tar file. Like same option in GNU tar.
numeric_owner = False

# Do no restore the uid/gid when finished, useful if you're restoring
# data without having root privileges or Unix users support
restore_ownership = True

# Whether to use plain HTTP (without SSL) to send data to S3
# See <https://bugs.launchpad.net/duplicity/+bug/433970>.
s3_unencrypted_connection = False

# Whether to use S3 Reduced Redudancy Storage
s3_use_rrs = False

# Whether to use S3 Infrequent Access Storage
s3_use_ia = False

# Whether to use S3 Glacier Storage
s3_use_glacier = False

# Whether to use S3 Glacier IR Storage
s3_use_glacier_ir = False

# Whether to use S3 Glacier Deep Archive Storage
s3_use_deep_archive = False

# Whether to use S3 One Zone Infrequent Access Storage
s3_use_onezone_ia = False

# Chunk size used for S3 multipart uploads.The number of parallel uploads to
# S3 be given by chunk size / volume size. Use this to maximize the use of
# your bandwidth. Defaults to 25MB
s3_multipart_chunk_size = 20 * 1024 * 1024

# Minimum chunk size accepted by S3
s3_multipart_minimum_chunk_size = 5 * 1024 * 1024

# Maximum number of processes to use while doing a multipart upload to S3
s3_multipart_max_procs = 4

# Use server side encryption in s3
s3_use_sse = False

# Use server side kms encryption in s3
s3_use_sse_kms = False
s3_kms_key_id = None
s3_kms_grant = None

# region and endpoint of s3
s3_region_name = None
s3_endpoint_url = None

# Which storage policy to use for Swift containers
swift_storage_policy = ""

# The largest size upload supported in a single put call for azure
azure_max_single_put_size = None

# The size of the blocks put to azure blob storage if bigger than azure_max_single_put_size
azure_max_block_size = None

# Maximum number of parallel connections to use when the blob size for azure exceeds 64MB
azure_max_connections = None

# Standard storage tier used for storring backup blobs (Hot|Cool|Archive).
azure_blob_tier = None

# Whether to use the full email address as the user name when
# logging into an imap server. If false just the user name
# part of the email address is used.
imap_full_address = False

# Name of the imap folder where we want to store backups.
# Can be changed with a command line argument.
imap_mailbox = "INBOX"

# Sync partial metadata by default
metadata_sync_mode = "partial"

# Wheter to specify --use-agent in GnuPG options
use_agent = False

# ssh duplicity_commands to use, used by ssh_pexpect (defaults to sftp, scp)
scp_command = None
sftp_command = None

# default to batch mode using public-key encryption
ssh_askpass = False

# user added ssh options
ssh_options = ""

# default cf backend is pyrax
cf_backend = "pyrax"

# default to fully deleting files in b2
b2_hide_files = False

# HTTPS ssl options (currently only webdav, lftp)
ssl_cacert_file = None
ssl_cacert_path = None
ssl_no_check_certificate = False

# user added rsync options
rsync_options = ""

# will be a Restart object if restarting
restart = None

# ignore (some) errors during operations; supposed to make it more
# likely that you are able to restore data under problematic
# circumstances. the default should absolutely always be True unless
# you know what you are doing.
ignore_errors = False

# Renames (--rename)
rename = {}

# enable data comparison on verify runs
compare_data = False

# sequencial backend tasks by default
concurrency = 0

# When symlinks are encountered, the item they point to is copied rather than
# the symlink.
copy_links = False

# When selected, triggers a dry-run before a full or incremental to compute
# changes, then runs the real operation and keeps track of the real progress
progress = False

# Controls the upload progress messages refresh rate. Default: update each
# 3 seconds
progress_rate = 3

# Level of Redundancy in % for Par2 files
par2_redundancy = 10

# Verbatim par2 other options
par2_options = ""

# Number of par2 volumes
par2_volumes = 1

# If set, log the chnages is the set instead of the normal collection status
show_changes_in_set = None

# If set, collect only the file status, not the whole root.
file_changed = None

# If set collect the files_changed list in statistics
files_changed = True

# delay (in seconds) before next operation after failure
backend_retry_delay = 30

# option for mediafire to purge files on delete instead of sending to trash
mf_purge = False

# Fake root directory path for iDrived backend
idr_fakeroot = None

# whether to check remote metadata (requires private key)
check_remote = True

# log verbosity.
verbosity = log.NOTICE

# whether 'inc` is explicit or not
# inc_explicit = True

# used in testing only -- set current time
current_time = None

# used in testing only - raises exception after volume
fail_on_volume = 0

# used in testing only - skips uploading a particular volume
skip_volume = 0

# used in testing only - fail BackendWrapper.put() on difftar volN
put_fail_volume = 0


# default filesystem encoding
# It seems that sys.getfilesystemencoding() will normally return
# 'utf-8' or some other sane encoding, but will sometimes fail and returns
# either 'ascii' or None.  Both are bogus, so default to 'utf-8' if it does.
fsencoding = sys.getfilesystemencoding()
fsencoding = fsencoding if fsencoding not in ["ascii", "ANSI_X3.4-1968", None] else "utf-8"


def dump_dict(config):
    """
    Return the "pickleable" configuration as a dictionary.

    Skips all internal attributes (e.g. ``__var__``) and attributes
    where pickling fails.

    Known skipped attributes:
      - ``select``: ``duplicity.selection.Select`` object
      - ``lockfile``: ``fasteners.process_lock.InterProcessLock`` object

    :param config: The configuration module or object to dump.
    :returns: A dictionary containing the pickleable configuration.
    :rtype: dict
    """
    c = {}
    for k, v in config.__dict__.items():
        try:
            if k.startswith("__") and k.endswith("__"):
                continue
            pickle.dumps(v)
        except (pickle.PicklingError, TypeError, AttributeError) as e:
            log.Debug(f"Skip {k}: {v} in config dump")
        else:
            c[k] = v
    return c


def load_dict(config_dict, config):
    """
    Update the configuration from a dictionary.

    :param dict config_dict: The dictionary containing configuration settings.
    :param config: The configuration module or object to update.
    """
    for k, v in config_dict.items():
        setattr(config, k, v)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2013 Matthieu Huin
# Copyright 2017 Xavier Lucas
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
OVH PCA (Public Cloud Archive) backend.
"""

import os
import time

import duplicity.backend
from duplicity import errors
from duplicity import log
from duplicity.errors import BackendException


class PCABackend(duplicity.backend.Backend):
    """
    Backend for OVH PCA
    """

    def __init__(self, parsed_url):
        """
        Initialize the PCA backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        try:
            from swiftclient import Connection
            from swiftclient import ClientException
        except ImportError as e:
            raise BackendException(f"""PCA backend requires the python-swiftclient library.
Exception: {str(e)}""")

        self.resp_exc = ClientException
        self.conn_cls = Connection
        conn_kwargs = {}

        # if the user has already authenticated
        if "PCA_PREAUTHURL" in os.environ and "PCA_PREAUTHTOKEN" in os.environ:
            conn_kwargs["preauthurl"] = os.environ["PCA_PREAUTHURL"]
            conn_kwargs["preauthtoken"] = os.environ["PCA_PREAUTHTOKEN"]

        else:
            if "PCA_USERNAME" not in os.environ:
                raise BackendException("PCA_USERNAME environment variable " "not set.")

            if "PCA_PASSWORD" not in os.environ:
                raise BackendException("PCA_PASSWORD environment variable " "not set.")

            if "PCA_AUTHURL" not in os.environ:
                raise BackendException("PCA_AUTHURL environment variable " "not set.")

            conn_kwargs["user"] = os.environ["PCA_USERNAME"]
            conn_kwargs["key"] = os.environ["PCA_PASSWORD"]
            conn_kwargs["authurl"] = os.environ["PCA_AUTHURL"]

        os_options = {}

        if "PCA_AUTHVERSION" in os.environ:
            conn_kwargs["auth_version"] = os.environ["PCA_AUTHVERSION"]
            if os.environ["PCA_AUTHVERSION"] == "3":
                if "PCA_USER_DOMAIN_NAME" in os.environ:
                    os_options.update({"user_domain_name": os.environ["PCA_USER_DOMAIN_NAME"]})
                if "PCA_USER_DOMAIN_ID" in os.environ:
                    os_options.update({"user_domain_id": os.environ["PCA_USER_DOMAIN_ID"]})
                if "PCA_PROJECT_DOMAIN_NAME" in os.environ:
                    os_options.update({"project_domain_name": os.environ["PCA_PROJECT_DOMAIN_NAME"]})
                if "PCA_PROJECT_DOMAIN_ID" in os.environ:
                    os_options.update({"project_domain_id": os.environ["PCA_PROJECT_DOMAIN_ID"]})
                if "PCA_TENANTNAME" in os.environ:
                    os_options.update({"tenant_name": os.environ["PCA_TENANTNAME"]})
                if "PCA_ENDPOINT_TYPE" in os.environ:
                    os_options.update({"endpoint_type": os.environ["PCA_ENDPOINT_TYPE"]})
                if "PCA_USERID" in os.environ:
                    os_options.update({"user_id": os.environ["PCA_USERID"]})
                if "PCA_TENANTID" in os.environ:
                    os_options.update({"tenant_id": os.environ["PCA_TENANTID"]})
                if "PCA_REGIONNAME" in os.environ:
                    os_options.update({"region_name": os.environ["PCA_REGIONNAME"]})

        else:
            conn_kwargs["auth_version"] = "2"
        if "PCA_TENANTNAME" in os.environ:
            conn_kwargs["tenant_name"] = os.environ["PCA_TENANTNAME"]
        if "PCA_REGIONNAME" in os.environ:
            os_options.update({"region_name": os.environ["PCA_REGIONNAME"]})

        conn_kwargs["os_options"] = os_options
        conn_kwargs["retries"] = 0

        self.conn_kwargs = conn_kwargs

        # This folds the null prefix and all null parts, which means that:
        #  //MyContainer/ and //MyContainer are equivalent.
        #  //MyContainer//My/Prefix/ and //MyContainer/My/Prefix are equivalent.
        url_parts = [x for x in parsed_url.path.split("/") if x != ""]

        self.container = url_parts.pop(0)
        if url_parts:
            self.prefix = f"{'/'.join(url_parts)}/"
        else:
            self.prefix = ""

        policy = "PCA"
        policy_header = "X-Storage-Policy"

        container_metadata = None
        try:
            self.conn = Connection(**self.conn_kwargs)
            container_metadata = self.conn.head_container(self.container)
        except ClientException:
            pass
        except Exception as e:
            raise errors.FatalError(
                f"Connection failed: {e.__class__.__name__} {str(e)}",
                log.ErrorCode.connection_failed,
            )

        if container_metadata is None:
            log.Info(f"Creating container {self.container}")
            try:
                headers = dict([[policy_header, policy]])
                self.conn.put_container(self.container, headers=headers)
            except Exception as e:
                raise errors.FatalError(
                    f"Container creation failed: {e.__class__.__name__} {str(e)}",
                    log.ErrorCode.connection_failed,
                )
        elif policy and container_metadata[policy_header.lower()] != policy:
            raise errors.FatalError(
                f"Container '{self.container}' exists but its storage policy is "
                f"'{container_metadata[policy_header.lower()]}' not '{policy}'."
            )

    def _error_code(self, operation, e):  # pylint: disable= unused-argument
        if isinstance(e, self.resp_exc):
            if e.http_status == 404:
                return log.ErrorCode.backend_not_found

    def _put(self, source_path, remote_filename):
        self.conn.put_object(
            self.container,
            self.prefix + os.fsdecode(remote_filename),
            open(os.fsdecode(source_path.name), "rb"),
        )

    def _get(self, remote_filename, local_path):
        body = self.unseal(self.prefix + os.fsdecode(remote_filename))
        if body:
            with open(os.fsdecode(local_path.name), "wb") as f:
                for chunk in body:
                    f.write(chunk)

    def __list_objs(self, ffilter=None):
        # full_listing should be set to True but a bug in python-swiftclient
        # doesn't forward query_string in this case...
        # bypass here using a patched copy (with query_string) from swiftclient code
        rv = self.conn.get_container(
            self.container,
            full_listing=False,
            path=self.prefix,
            query_string="policy_extra=true",
        )
        listing = rv[1]
        while listing:
            marker = listing[-1]["name"]
            version_marker = listing[-1].get("version_id")
            listing = self.conn.get_container(
                self.container,
                marker=marker,
                version_marker=version_marker,
                full_listing=False,
                path=self.prefix,
                query_string="policy_extra=true",
            )[1]
            if listing:
                rv[1].extend(listing)
        if ffilter is not None:
            return list(filter(ffilter, rv[1]))
        return rv[1]

    def _list(self):
        return [os.fsencode(o["name"][len(self.prefix) :]) for o in self.__list_objs()]

    def _delete(self, filename):
        self.conn.delete_object(self.container, self.prefix + os.fsdecode(filename))

    def _query(self, filename):
        sobject = self.conn.head_object(self.container, self.prefix + os.fsdecode(filename))
        return {"size": int(sobject["content-length"])}

    def unseal(self, remote_filename):
        try:
            _, body = self.conn.get_object(self.container, remote_filename, resp_chunk_size=1024)
            log.Info(f"File {remote_filename} was successfully unsealed.")
            return body
        except self.resp_exc as e:
            # The object is sealed but being released.
            if e.http_status == 429:
                # The retry-after header contains the remaining duration before
                # the unsealing operation completes.
                duration = int(e.http_response_headers["Retry-After"])
                m, s = divmod(duration, 60)
                h, m = divmod(m, 60)
                eta = f"{int(h)}h{int(m):02}m{int(s):02}s"
                log.Info(f"File {remote_filename} is being unsealed, operation ETA is {eta}.")
            else:
                raise errors.FatalError(
                    f"Connection failed: {e.__class__.__name__} {str(e)}",
                    log.ErrorCode.connection_failed,
                )
        return None

    def pre_process_download_batch(self, remote_filenames):
        """
        This is called before downloading volumes from this backend
        by main engine. For PCA, volumes passed as argument need to be unsealed.
        This method is blocking, showing a status at regular interval
        """
        retry_interval = 60  # status will be shown every 60s
        # remote_filenames are bytes string
        u_remote_filenames = list(map(os.fsdecode, remote_filenames))
        objs = self.__list_objs(
            ffilter=lambda x: os.fsdecode(x["name"]) in [self.prefix + s for s in u_remote_filenames]
        )
        # first step: retrieve pca seal status for all required volumes
        # and launch unseal for all sealed files
        one_object_not_unsealed = False
        for o in objs:
            filename = os.fsdecode(o["name"])
            # see ovh documentation for policy_retrieval_state definition
            policy_retrieval_state = o["policy_retrieval_state"]
            log.Info(f"Volume {filename}. State : {policy_retrieval_state}. ")
            if policy_retrieval_state == "sealed":
                log.Notice(f"Launching unseal of volume {filename}.")
                self.unseal(o["name"])
                one_object_not_unsealed = True
            elif policy_retrieval_state == "unsealing":
                one_object_not_unsealed = True
        # second step: display estimated time for last volume unseal
        # and loop until all volumes are unsealed
        while one_object_not_unsealed:
            one_object_not_unsealed = self.unseal_status(u_remote_filenames)
            time.sleep(retry_interval)
            # might be a good idea to show a progress bar here...
        else:
            log.Notice("All volumes to download are unsealed.")

    def unseal_status(self, u_remote_filenames):
        """
        Shows unsealing status for input volumes
        """
        one_object_not_unsealed = False
        objs = self.__list_objs(
            ffilter=lambda x: os.fsdecode(x["name"]) in [self.prefix + s for s in u_remote_filenames]
        )
        max_duration = 0
        for o in objs:
            policy_retrieval_state = o["policy_retrieval_state"]
            filename = os.fsdecode(o["name"])
            if policy_retrieval_state == "sealed":
                log.Notice(f"Error: volume is still in sealed state : {filename}.")
                log.Notice(f"Launching unseal of volume {filename}.")
                self.unseal(o["name"])
                one_object_not_unsealed = True
            elif policy_retrieval_state == "unsealing":
                duration = int(o["policy_retrieval_delay"])
                log.Info(f"{filename} available in {int(duration)} seconds.")
                if duration > max_duration:
                    max_duration = duration
                one_object_not_unsealed = True

        m, s = divmod(max_duration, 60)
        h, m = divmod(m, 60)
        max_duration_eta = f"{int(h)}h{int(m):02}m{int(s):02}s"
        log.Notice(f"Need to wait {max_duration_eta} before all volumes are unsealed.")
        return one_object_not_unsealed


duplicity.backend.register_backend("pca", PCABackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
# Copyright 2014 Aaron Whitehouse
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Functions for matching shell globs.

All functions in this module only accept unicode. Any byte strings should
be converted to unicode before sending them into this.
"""

import re


class GlobbingError(Exception):
    """
    Something has gone wrong when parsing a glob string
    """

    pass


class FilePrefixError(GlobbingError):
    """
    Signals that a specified file doesn't start with correct prefix
    """

    pass


def _glob_get_prefix_regexs(glob_str):
    """
    Return list of regexps equivalent to prefixes of glob_str
    """
    # Internal. Used by glob_get_sf.
    glob_parts = glob_str.split("/")
    if "" in glob_parts[1:-1]:
        # "" OK if comes first or last, as in /foo/
        raise GlobbingError(f"Consecutive '/'s found in globbing string {glob_str}")

    prefixes = ["/".join(glob_parts[: i + 1]) for i in range(len(glob_parts))]
    # we must make exception for root "/", only dir to end in slash
    if prefixes[0] == "":
        prefixes[0] = "/"
    return list(map(glob_to_regex, prefixes))


def select_fn_from_glob(glob_str, include, ignore_case=False):
    """
    Return a selection function based on the given glob string.

    The returned function ``test_fn(path)`` tests whether a path matches the
    glob, following Unix shell rules.

    :param str glob_str: The glob string to match against.
    :param int include: 1 if the glob is an include glob, 0 if it is an exclude glob.
    :param bool ignore_case: Whether to ignore case when matching.
    :returns: A selection function that returns:
        - 0 if the file should be excluded
        - 1 if the file should be included
        - 2 if the folder should be scanned
        - None if the test has nothing to say about the file
    :rtype: callable
    """
    assert isinstance(glob_str, str), f"glob_str must be a str (unicode), got {type(glob_str).__name__}: {glob_str!r}"
    glob_ends_w_slash = False

    if glob_str == "/":
        # If the glob string is '/', it implicitly includes everything
        glob_str = "/**"
    elif glob_str[-1] == "/":
        glob_ends_w_slash = True
        # Remove trailing / from directory name (unless that is the entire
        # string)
        glob_str = glob_str[:-1]

    flags = 0
    if ignore_case:
        flags = re.IGNORECASE

    re_comp = lambda r: re.compile(r, re.S | flags)

    # matches what glob matches and any files in directory
    # Resulting regular expression is:
    # ^ string must be at the beginning of path
    # string translated into regex
    # ($|/) nothing must follow except for the end of the string, newline or /
    # Note that the "/" at the end of the regex means that it will match
    # if the glob matches a parent folders of path, i.e. including a folder
    # includes everything within it.
    glob_comp_re = re_comp(f"^{glob_to_regex(glob_str)}($|/)")

    if glob_ends_w_slash:
        # Creates a version of glob_comp_re that does not match folder contents
        # This can be used later to check that an exact match is actually a
        # folder, rather than a file.
        glob_comp_re_exact = re_comp(f"^{glob_to_regex(glob_str)}($)")

    if glob_str.find("**") != -1:
        # glob_str has a ** in it
        glob_str = glob_str[: glob_str.find("**") + 2]  # truncate after **

    # Below regex is translates to:
    # ^ string must be at the beginning of path
    # the regexs corresponding to the parent directories of glob_str
    # $ nothing must follow except for the end of the string or newline
    scan_comp_re = re_comp(f"^({'|'.join(_glob_get_prefix_regexs(glob_str))})$")

    def test_fn(path):
        assert not path.uc_name[-1] == "/" or path.uc_name == "/", (
            "path.name should never end in '/' during normal operation for "
            "normal paths (except '/' alone)\n"
            "path.name here is " + path.uc_name + " and glob is " + glob_str
        )

        if glob_comp_re.match(path.uc_name):
            # Path matches glob, or is contained within a matching folder
            if not glob_ends_w_slash:
                return include
            else:
                # Glob ended with a /, so we need to check any exact match was
                # a folder
                if glob_comp_re_exact.match(path.uc_name):
                    # Not an included file/folder, so must be a folder to match
                    if path.isdir():
                        # Is a directory, so all is well
                        return include
                    else:
                        # Exact match and not a folder
                        return None
                else:
                    # An included file/folder, so normal approach is fine
                    return include

        elif include == 1 and scan_comp_re.match(path.uc_name):
            return 2
        else:
            return None

    return test_fn


def glob_to_regex(pat):
    """
    Convert a shell glob pattern to a regular expression.

    Currently only the ``?``, ``*``, ``[]``, and ``**`` expressions are supported.
    Ranges like ``[a-z]`` are currently unsupported. There is no way to
    quote these special characters.

    :param str pat: The shell glob pattern.
    :returns: An equivalent regular expression string.
    :rtype: str
    """
    # Internal. Used by glob_get_sf, glob_get_prefix_res and unit tests.

    assert isinstance(pat, str), f"pat must be a str (unicode), got {type(pat).__name__}: {pat!r}"

    i, n, res = 0, len(pat), ""
    while i < n:
        c, s = pat[i], pat[i : i + 2]
        i = i + 1
        if s == "**":
            res = f"{res}.*"
            i = i + 1
        elif c == "*":
            res = f"{res}[^/]*"
        elif c == "?":
            res = f"{res}[^/]"
        elif c == "[":
            j = i
            if j < n and pat[j] in "!^":
                j = j + 1
            if j < n and pat[j] == "]":
                j = j + 1
            while j < n and pat[j] != "]":
                j = j + 1
            if j >= n:
                res = f"{res}\\["  # interpret the [ literally
            else:
                # Deal with inside of [..]
                stuff = pat[i:j].replace("\\", "\\\\")
                i = j + 1
                if stuff[0] in "!^":
                    stuff = f"^{stuff[1:]}"
                res = f"{res}[{stuff}]"
        else:
            res = res + re.escape(c)
    return res



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Backend pool for running backend commands in parallel.
"""

import concurrent.futures
import multiprocessing
import multiprocessing.connection
import os
import sys
import time
import traceback
from collections import Counter
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional

from duplicity import backend
from duplicity import config
from duplicity import file_naming
from duplicity import errors
from duplicity import log
from duplicity import path
from duplicity import util

pool_backend = None


@dataclass
class TrackRecord:
    """
    Result tracking for commands executed in the pool.
    """

    # result tracking while executed in the pool
    track_id: int
    pid: int
    trace_back: Optional[List[str]] = field(
        default_factory=list
    )  # trace backs can't be pickled, store as string, to get it over into main process
    result: object = None  # must be picklable
    log_prefix: str = ""
    start_time: datetime = field(default=datetime.now())
    stop_time: datetime = field(default=datetime.min)

    def __post_init__(self):
        self.start_time = datetime.now()

    def get_runtime(self) -> timedelta:
        """
        Return the runtime of the command as a timedelta.
        """
        return self.stop_time - self.start_time


def track_cmd(track_id, cmd_name: str, *args, **kwargs):
    """
    wraps the pooled function for time tracking and exception handling.
    Recording the trace back of an exception when still in process pool context
    to point to the right place.
    (This function can't be part of the BackendPool, as then the whole class get pickled)
    """
    global pool_backend
    p = multiprocessing.current_process()
    trk_rcd = TrackRecord(track_id, p.pid, log_prefix=log.PREFIX)  # type: ignore
    # send cmd/process assignment back to pool for tracking.
    try:
        cmd = getattr(pool_backend, cmd_name)
        trk_rcd.result = cmd(*args, **kwargs)
    except Exception as e:
        trk_rcd.result = e
        trk_rcd.trace_back = traceback.format_tb(e.__traceback__)
    trk_rcd.stop_time = datetime.now()
    log.Debug(f"Command done: {trk_rcd} ")
    return trk_rcd


class BackendPool:
    """
    uses concurrent.futures.ProcessPoolExecutor to run backend commands in background
    """

    @dataclass
    class CmdStatus:
        """
        Status of a command in the pool.
        """

        function_name: str
        args: Dict
        kwargs: List
        trk_rcd: Optional[TrackRecord] = None
        done: bool = False

    def __init__(self, url_string, processes=None) -> None:
        config_str = config.dump_dict(config)
        self.ppe = concurrent.futures.ProcessPoolExecutor(
            max_workers=processes,
            initializer=self._process_init,
            initargs=(url_string, config_str),
            mp_context=multiprocessing.get_context(method="spawn"),
        )
        self._command_queue = []
        self._track_id = 0
        self.all_results: List[TrackRecord] = []
        self._all_futures: List[concurrent.futures.Future] = []
        self.peak_in_queue = 0
        self._command_throttle_stats = []
        self._last_results_reported = 0

    def __enter__(self):
        return self.ppe

    def __exit__(self, type, value, traceback):
        self.shutdown()

    @staticmethod
    def _process_init(url_string, config_dump):
        global pool_backend
        config.load_dict(config_dump, config)
        pid = os.getpid()
        pool_nr = multiprocessing.current_process()._identity[0]
        # get the multiprocessing logger with stderr stream handler added
        log.setup()
        log.PREFIX = f"Pool{pool_nr}: "
        log.setverbosity(config.verbosity)
        if config.verbosity == log.DEBUG:
            logger = multiprocessing.log_to_stderr()
        log.Info(f"Staring pool process with pid: {pid}")
        file_naming.prepare_regex()
        # util.start_debugger()
        backend.import_backends()
        pool_backend = backend.get_backend(url_string)

    def command(self, func_name, args=(), kwds=None):
        """
        Run a function in a pool of independent processes.

        Call the function by name.

        :param str func_name: The name of the backend method to execute.
        :param tuple args: Positional arguments for the method.
        :param dict kwds: Keyword arguments for the method.
        :returns: A unique ID for the command.
        :rtype: int
        """
        if kwds is None:
            kwds = {}
        self._track_id += 1
        self._command_queue.append(
            self.ppe.submit(
                track_cmd,
                self._track_id,
                func_name,
                *args,
            ),
        )
        self._collect_finished_cmds()
        return self._track_id

    def _raise_exception_if_any(self, track_rcrd):
        """
        Raise the exception stored in ``track_rcrd`` if it occurred while running the command.

        :param TrackRecord track_rcrd: The track record to check for exceptions.
        :raises FatalError: If the exception has a ``.code`` attribute.
        :raises Exception: The original exception if it doesn't have a ``.code`` attribute.
        """
        if isinstance(track_rcrd.result, Exception):
            exception_str = f"{''.join(track_rcrd.trace_back)}\n{track_rcrd.result}"
            log.Debug(f"Exception thrown in pool: \n{exception_str}")
            if hasattr(track_rcrd.result, "code"):
                raise errors.FatalError(
                    f"Exception {track_rcrd.result.__class__.__name__} in background "
                    f"pool {track_rcrd.log_prefix}. "
                    "For trace back set loglevel to DEBUG and check output for given pool.",
                    code=track_rcrd.result.code,  # type: ignore
                )
            else:
                raise track_rcrd.result

    def _collect_finished_cmds(self):
        """
        store finished results in `all_results` and remove command from queue
        """
        try:
            finished_commands = concurrent.futures.as_completed(self._command_queue, timeout=0.1)
            for result in finished_commands:
                track_rcrd = result.result(timeout=0)
                self._raise_exception_if_any(track_rcrd)
                self.all_results.append(track_rcrd)
                self._all_futures.append(result)
                self._command_queue.remove(result)
                self.peak_in_queue = max(self.peak_in_queue, len(self._command_queue))
        except concurrent.futures._base.TimeoutError:
            pass

    def command_throttled(self, func_name, commands_in_buffer=1, args=(), kwds=None):
        """
        Run a function in the pool, blocking if the queue is full.

        Blocks if the queue gets larger than the number of workers plus
        ``commands_in_buffer``.

        :param str func_name: The name of the backend method to execute.
        :param int commands_in_buffer: Number of extra commands to allow in the buffer.
        :param tuple args: Positional arguments for the method.
        :param dict kwds: Keyword arguments for the method.
        :returns: A unique ID for the command.
        :rtype: int
        """

        if kwds is None:
            kwds = {}

        # define accepted queue buffer.
        max_pending_queue_len = self.ppe._max_workers + commands_in_buffer  # type: ignore

        start = datetime.now()
        if len(self._command_queue) >= max_pending_queue_len:
            # queue full wait for free slot
            finished_commands = concurrent.futures.as_completed(self._command_queue)
            next(finished_commands)  # blocking until next finished

        self._command_throttle_stats.append((datetime.now() - start).total_seconds())
        log.Debug(f"Command delayed by {self._command_throttle_stats[-1]:.2f}")
        # queue command
        track_id = self.command(func_name, args, kwds)

        return track_id

    def get_queue_length(self) -> int:
        """
        Return the number of commands currently in the queue.
        """
        return len(self._command_queue)

    def results_since_last_call(self, from_pos=None) -> Iterator[TrackRecord]:
        """
        Collect results from commands finished since the last run of this method.

        This method should be called from one receiver only, as it keeps track
        of what was sent internally.

        :param int from_pos: Use to overwrite the internal pointer ``_last_results_reported``.
        :returns: An iterator of track records.
        :rtype: Iterator[TrackRecord]
        """
        self._collect_finished_cmds()
        if from_pos is not None:
            self._last_results_reported = from_pos
        end = len(self.all_results)
        for result in self.all_results[self._last_results_reported : end]:
            yield result
            self._last_results_reported = end

    def get_stats(self, last_index=None):
        """
        Return statistics about the pool execution.

        :param int last_index: If provided, only consider results up to this index.
        :returns: A dictionary containing statistics.
        :rtype: dict
        """
        vals = [x.get_runtime().total_seconds() for x in self.all_results[:last_index]]
        count = len(vals)
        if count > 0:
            avg_time = sum(vals) / count
            max_time = max(vals)
            min_time = min(vals)
            if len(self._command_throttle_stats) > 0:
                avg_throttle = sum(self._command_throttle_stats) / len(self._command_throttle_stats)
                max_throttle = max(self._command_throttle_stats)
                min_throttle = min(self._command_throttle_stats)
            else:
                avg_throttle = max_throttle = min_throttle = -1
        else:
            avg_time = max_time = min_time = avg_throttle = max_throttle = min_throttle = -1
        pool_usage = Counter([x.log_prefix for x in self.all_results[:last_index]])

        log.Debug(
            f"count: {count}, avg: {avg_time}, max: {max_time}, min: {min_time}, pool usage: {pool_usage}, "
            f"peak in queue: {self.peak_in_queue}, avg throttle: {avg_throttle}"
        )
        return {
            "count": count,
            "time": {"avg": avg_time, "max": max_time, "min": min_time},
            "throttle": {"avg": avg_throttle, "max": max_throttle, "min": min_throttle},
            "pool_usage": pool_usage,
            "peak_in_queue": self.peak_in_queue,
        }

    def shutdown(self, wait=True, cancel_futures=True):
        """
        Shutdown the process pool.

        :param bool wait: Whether to wait for all processes to finish.
        :param bool cancel_futures: Whether to cancel any pending futures.
        """
        log.Debug("Process Pool: Start shutdown.")
        if sys.version_info >= (3, 9):
            self.ppe.shutdown(wait=wait, cancel_futures=cancel_futures)
        else:
            self.ppe.shutdown(wait=wait)
        log.Debug("Process Pool: Shutdown done.")


# code to run/test the pool independent, not relevant for duplicity
if __name__ == "__main__":
    from duplicity import config, log

    log.setup()
    log.add_file("/tmp/tmp.log")
    config.verbosity = log.INFO
    log.setverbosity(config.verbosity)
    backend.import_backends()
    config.async_concurrency = 4
    config.num_retries = 2
    url = "file:///tmp/test_direct"
    bw = backend.get_backend(url)
    # ^^^^^^^^^^ above commands are only there for mocking a duplicity config

    start_time = time.time()
    bpw = BackendPool(url, processes=config.async_concurrency)
    results: List[TrackRecord] = []
    with bpw as pool:
        # issue tasks into the process pool
        import pathlib

        if len(sys.argv) > 1:
            src = sys.argv[1]
        else:
            src = "./"
        for file in [file for file in pathlib.Path(src).iterdir() if file.is_file()]:
            source_path = path.Path(file.as_posix())
            bpw.command(bw.put_validated.__name__, args=(source_path, source_path.get_filename()))  # type: ignore
            cmd_results = [x for x in bpw.results_since_last_call()]
            log.Info(f"got: {len(cmd_results)}, cmd left: {bpw.get_queue_length()}, track_id: {bpw._track_id}")
            results.extend(cmd_results)

        # wait for tasks to complete
        suppress_log = False
        while True:
            cmd_results = [x for x in bpw.results_since_last_call()]
            if len(cmd_results) > 0 or not suppress_log:
                log.Info(
                    f"got: {len(cmd_results)}, {not suppress_log}, "
                    f"precessed {len(results)} cmd left: {bpw.get_queue_length()}, track_id: {bpw._track_id}"
                )
                suppress_log = False
            if len(cmd_results) == 0:
                suppress_log = True

            results.extend(cmd_results)
            if bpw.get_queue_length() == 0:
                break

        bpw.get_stats(last_index=-1)
        input("Press Enter to continue...")
    # process pool is closed automatically

    log.Notice(f"Bytes written: {sum([int(x.result) for x in results])}")  # type: ignore
    log.Notice(f"Time elapsed: {time.time() - start_time}")
